"""
Markov state models are a statistical tool for analysing molecular simulations which has met with lots of success.
The Model class here, encapsulates all functionallity for the calculation of Markov models while hiding unnecessary
details under the hood. It uses PyEMMA [1] internally to calculate Markov models.

References
----------
.. [1] PyEMMA 2: A Software Package for Estimation, Validation, and Analysis of Markov Models. Martin K. Scherer et al. JCTC 2015.
"""
# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
import numpy as np
from scipy import stats
import warnings
import random
from moleculekit.molecule import Molecule
from moleculekit.vmdviewer import getCurrentViewer
from htmd.units import convert as unitconvert
import logging
logger = logging.getLogger(__name__)


class Model(object):
    """ Constructor for the Model class.

    Parameters
    ----------
    data : :class:`MetricData <htmd.metricdata.MetricData>` object
        A :class:`MetricData <htmd.metricdata.MetricData>` object containing the discretized trajectories

    Example
    -------
    >>> model = Model(data)

    .. rubric:: Methods
    .. autoautosummary:: htmd.model.Model
        :methods:
    .. rubric:: Attributes
    .. autoautosummary:: htmd.model.Model
        :attributes:
    """

    def __init__(self, data=None, file=None):
        if data is None:
            if file is not None:
                self.load(file)
            return
        else:
            self.data = data
        self.hmm = None
        self._modelid = None
        if self.data._clusterid is None:
            raise NameError('You need to cluster your data before making a Markov model')
        if self.data._dataid != self.data._clusterid:
            raise NameError('You have modified the data in data.dat after clustering. Please re-cluster.')
        self._clusterid = self.data._clusterid

    def markovModel(self, lag, macronum, units='frames', sparse=False, hmm=False):
        """ Build a Markov model at a given lag time and calculate metastable states

        Parameters
        ----------
        lag : int
            The lag time at which to calculate the Markov state model. The units are specified with the `units` argument.
        macronum : int
            The number of macrostates (metastable states) to produce
        units : str
            The units of lag. Can be 'frames' or any time unit given as a string.
        sparse : bool
            Make the transition matrix sparse. Useful if lots (> 4000) states are used for the MSM. Warning: untested.

        Examples
        --------
        >>> model = Model(data)
        >>> model.markovModel(150, 4)  # 150 frames lag, 4 macrostates
        """
        import pyemma.msm as msm
        self._integrityCheck(markov=True)

        lag = unitconvert(units, 'frames', lag, fstep=self.data.fstep)

        self.lag = lag
        self.msm = msm.estimate_markov_model(self.data.St.tolist(), self.lag, sparse=sparse)
        modelflag = False
        while not modelflag:
            self.coarsemsm = self.msm.pcca(macronum)
            if len(np.unique(self.msm.metastable_assignments)) != macronum:
                macronum -= 1
                logger.warning('PCCA returned empty macrostates. Reducing the number of macrostates to {}.'.format(macronum))
            else:
                modelflag = True
            if macronum < 2:
                raise RuntimeError('Could not create even two macrostates. Please revise your clustering.')

        self._modelid = random.random()

        if hmm:  # Still in development
            self.hmm = self.msm.coarse_grain(self.macronum)

        logger.info('{:.1f}% of the data was used'.format(self.msm.active_count_fraction * 100))

        _macroTrajectoriesReport(self.macronum, _macroTrajSt(self.data.St, self.macro_ofcluster), self.data.simlist)

    def createState(self, microstates=None, indexpairs=None):
        """ Creates a new state. Works both for new clusters and macrostates.

        If creating a new cluster, it just reassigns the given frames to the new cluster.
        If creating a new macrostate, it removes the given microstates from their previous macrostate, creates a new one
        and assigns them to it.

        Parameters
        ----------
        microstates : list
            The microstates to split out into a new macrostate.
        indexpairs : list
            List of lists. Each row is a simulation index-frame pair which should be added to a new cluster.
        """
        if microstates is not None and indexpairs is not None:
            raise AttributeError('microstates and indexpairs arguments are mutually exclusive')
        if microstates is not None:
            newmacro = self.macronum

            # Fixing sets. Remove microstates from previous macrostates and add new set
            for i, metset in enumerate(self.msm.metastable_sets):
                self.msm.metastable_sets[i] = np.array(np.sort(list(set(metset) - set(microstates))))
            self.msm.metastable_sets.append(np.array(microstates, dtype=np.int64))

            todelete = np.where([len(x) == 0 for x in self.msm.metastable_sets])[0]

            # Fixing hard assignments
            self.msm.metastable_assignments[microstates] = newmacro

            # Fixing memberships. Padding the array with 0s for the new macrostate
            self.msm._metastable_memberships = np.pad(self.msm.metastable_memberships, ((0, 0), (0, 1)), mode='constant', constant_values=(0))
            self.msm._metastable_memberships[microstates, :] = 0
            self.msm._metastable_memberships[microstates, -1] = 1

            # Moving probabilities of empty states to new one
            othermicro = np.ones(self.micronum, dtype=bool)
            othermicro[microstates] = False
            othermicro = np.where(othermicro)[0]
            self.msm._metastable_memberships[othermicro, -1] = np.sum(self.msm._metastable_memberships[othermicro[:, None], todelete], axis=1)

            # Fixing distributions
            self.msm._metastable_distributions = np.pad(self.msm.metastable_distributions, ((0, 1), (0, 0)), mode='constant', constant_values=(0))
            self.msm._metastable_distributions[-1, microstates] = 1 / len(microstates)
        if indexpairs is not None:
            newcluster = self.data.K
            for ip in indexpairs:
                self.data.trajectories[ip[0]].cluster[ip[1]] = newcluster
            self.data.K += 1
            self.data.N = np.bincount(np.concatenate(self.data.St))

    @property
    def P(self):
        """ The transition probability matrix """
        return self.msm.transition_matrix

    @property
    def micro_ofcluster(self):
        """ Mapping of clusters to microstates

        Numpy array which at index i has the index of the microstate corresponding to cluster i.
        Clusters which were not connected and thus are not in the model have a microstate value of -1.
        """
        self._integrityCheck(postmsm=True)
        micro_ofcluster = -np.ones(self.data.K, dtype=int)
        micro_ofcluster[self.msm.active_set] = np.arange(len(self.msm.active_set))
        return micro_ofcluster

    @property
    def cluster_ofmicro(self):
        """ Mapping of microstates to clusters

        Numpy array which at index i has the index of the cluster corresponding to microstate i.
        """
        self._integrityCheck(postmsm=True)
        return self.msm.active_set

    @property
    def micronum(self):
        """ Number of microstates """
        self._integrityCheck(postmsm=True)
        return len(self.msm.active_set)

    @property
    def macronum(self):
        """ Number of macrostates """
        self._integrityCheck(postmsm=True)
        return len(set(self.msm.metastable_assignments))

    @property
    def macro_ofmicro(self):
        """ Mapping of microstates to macrostates

        Numpy array which at index i has the index of the macrostate corresponding to microstate i.
        """
        self._integrityCheck(postmsm=True)
        # Fixing pyemma macrostate numbering
        mask = np.ones(np.max(self.msm.metastable_assignments) + 1, dtype=int) * -1
        mask[list(set(self.msm.metastable_assignments))] = range(self.macronum)
        return mask[self.msm.metastable_assignments]

    @property
    def macro_ofcluster(self):
        """ Mapping of clusters to microstates

        Numpy array which at index i has the index of the macrostate corresponding to cluster i.
        Clusters which were not connected and thus are not in the model have a macrostate value of -1.
        """
        self._integrityCheck(postmsm=True)
        macro_ofcluster = -np.ones(self.data.K, dtype=int)
        macro_ofcluster[self.msm.active_set] = self.macro_ofmicro
        return macro_ofcluster

    def plotTimescales(self, lags=None, minlag=None, maxlag=None, numlags=25, units='frames', errors=None, nits=None,
                       results=False, plot=True, save=None, njobs=-2):
        """ Plot the implied timescales of MSMs of various lag times

        Parameters
        ----------
        lags : list
            Specify explicitly at which lag times to compute the timescales.
        minlag: float
            The minimum lag time for the timescales. Used in combination with `maxlag` and `numlags`.
        maxlag: float
            The maximum lag time for the timescales. If None will default to the mode length of the trajectories.
        numlags: int
            The number of points to place between `minlag` and `maxlag`.
        units : str
            The units of lag. Can be 'frames' or any time unit given as a string.
        errors : errors
            Calculate errors using Bayes (Refer to pyEMMA documentation)
        nits : int
            Number of implied timescales to calculate. Default: all
        results : bool
            If the method should return the calculated implied timescales
        plot : bool
            If the method should display the plot of implied timescales
        save : str
            Path of the file in which to save the figure
        njobs : int
            Number of parallel jobs to spawn for calculation of timescales. Negative numbers are used for spawning jobs as many as CPU threads. 
            -1: for all CPUs -2: for all except one etc.

        Returns
        -------
        If given results=True this method will return the following data
        its : np.ndarray
            The calculated implied timescales. 2D array with dimensions (len(`lags`), `nits`)
        lags : np.ndarray
            A list of the lag times that were used to calculate the implied timescales

        Examples
        --------
        >>> model = Model(data)
        >>> model.plotTimescales()
        >>> model.plotTimescales(lags=list(range(1,100,5)))
        >>> model.plotTimescales(minlag=0.1, maxlag=20, numlags=25, units='ns')
        """
        import pyemma.plots as mplt
        import pyemma.msm as msm
        self._integrityCheck()
        if lags is None:
            lags = self.data._defaultLags(minlag, maxlag, numlags, units)
        else:
            lags = unitconvert(units, 'frames', lags, fstep=self.data.fstep).tolist()

        if nits is None:
            nits = np.min((self.data.K, 20))

        from htmd.config import _config
        its = msm.its(self.data.St.tolist(), lags=lags, errors=errors, nits=nits, n_jobs=njobs) # Use all CPUs minus one
        if plot or (save is not None):
            from matplotlib import pylab as plt
            plt.ion()
            plt.figure()
            try:
                mplt.plot_implied_timescales(its, dt=self.data.fstep, units='ns')
            except ValueError as ve:
                plt.close()
                raise ValueError('{} This is probably caused by badly set fstep in the data ({}). '.format(ve, self.data.fstep) +
                                 'Please correct the model.data.fstep to correspond to the simulation frame step in nanoseconds.')
            if save is not None:
                plt.savefig(save, dpi=300, bbox_inches='tight', pad_inches=0.2)
            if plot:
                plt.show()
        if results:
            return its.get_timescales(), its.lags

    def maxConnectedLag(self, lags):
        """ Heuristic for getting the lagtime before a timescale drops.

        It calculates the last lagtime before a drop occurs in the first implied timescale due to disconnected states.
        If the top timescale is closer to the second top timescale at the previous lagtime than to itself at the previous
        lagtime it means that a drop occured. The lagtime before the drop is returned.

        Parameters
        ----------
        lags : np.ndarray or list
            A list of lag times for which to calculate the implied timescales

        Returns
        -------
        ml : int
            The maximum lagtime before a drop occurs in the top timescale

        Examples
        --------
        >>> model = Model(data)
        >>> model.maxConnectedLag(list(range(1, 100, 5)))
        """
        if len(lags) == 1:
            return lags
        if isinstance(lags, np.ndarray):
            lags = lags.astype(int)

        import pyemma.msm as msm
        itime = msm.its(self.data.St.tolist(), lags=lags, nits=2).get_timescales()

        for i in range(1, np.size(itime, 0)):
            if abs(itime[i, 0] - itime[i-1, 1]) < abs(itime[i, 0] - itime[i-1, 0]):
                lagidx = i-1
                break
            else:
                lagidx = i
        return lags[lagidx], itime

    def sampleStates(self, states=None, frames=20, statetype='macro', replacement=False, samplemode='random', allframes=False):
        """ Samples frames from a set of states

        Parameters
        ----------
        states : Union[list, int]
            A list of state indexes from which we want to sample
        frames : Union[None, int, list]
            An integer with the number of frames we want to sample per state or a list of same length as
            `states` which contains the number of frames we want from each of the states.
            If set to None it will return all frames of the states.
        statetype : ['micro','macro','cluster'], optional
            The type of state we want to sample from.
        replacement : bool
            If we want to sample with or without replacement.
        samplemode : ['random','even','weighted']
            What sampling strategy to use. For `statetype` == 'macro' this can be set to 'even' to sample evenly from
            all microstates in the macrostate or to 'weighted' to sample proportional to the equilibium probability of
            each microstate in the macrostate.
        allframes : bool
            Deprecated. Use frames=None instead.

        Returns
        -------
        absframes : numpy.ndarray
            An array which contains for each state an array containing absolute trajectory frames
        relframes : numpy.ndarray
            An array which contains for each state a 2D array containing the trajectory ID and frame number for each of
            the sampled frames

        Examples
        --------
        >>> model = Model(data)
        >>> model.markovModel(100, 5)
        >>> model.sampleStates(range(5), [10, 3, 2, 50, 1])  # Sample from all 5 macrostates
        >>> model.sampleStates(range(model.micronum), samplesnum, statetype='micro')  # Sample from all microstates
        """
        if statetype == 'cluster':
            if samplemode != 'random':
                logger.warning("'cluster' states incompatible with 'samplemode' other than 'random'. Defaulting to 'random'")
            return self.data.sampleClusters(states, frames, replacement, allframes)

        if states is None:
            if statetype == 'macro':
                states = range(self.macronum)
            elif statetype == 'micro':
                states = range(self.micronum)
        if isinstance(states, int):
            states = [states, ]

        if allframes:
            logger.warning('The allframes option will be deprecated. Use frames=None instead.')
            frames = None

        self._integrityCheck(postmsm=True)
        if statetype != 'macro' and samplemode != 'random':
            samplemode = 'random'
            logger.warning("'micro' states incompatible with 'samplemode' other than 'random'. Defaulting to 'random'")

        if frames is None or isinstance(frames, int):
            frames = np.repeat(frames, len(states))

        stConcat = np.concatenate(self.data.St)
        absFrames = []
        relFrames = []
        for i in range(len(states)):
            if frames[i] == 0:
                absFrames.append(np.array([], dtype=int))
                relFrames.append(np.array([], dtype=int))
                continue

            st = states[i]
            if statetype == 'macro':
                (selFr, selMicro) = _sampleMacro(self, st, stConcat, samplemode, frames[i], replacement)
                absFrames.append(selFr)
            elif statetype == 'micro':
                absFrames.append(_sampleMicro(self, st, stConcat, frames[i], replacement))
            else:
                raise NameError('No valid state type given (read documentation)')

            if len(absFrames[-1]) == 0:
                raise NameError('No frames could be sampled from {} state {}. State is empty.'.format(statetype, st))

            relFrames.append(self.data.abs2rel(absFrames[-1]))
        return absFrames, relFrames

    def eqDistribution(self, plot=True, save=None):
        """ Obtain and plot the equilibrium probabilities of each macrostate

        Parameters
        ----------
        plot : bool, optional, default=True
            Disable plotting of the equilibrium distribution by setting it to False
        save : str
            Path of the file in which to save the figure

        Returns
        -------
        eq : ndarray
            An array of equilibrium probabilities of the macrostates

        Examples
        --------
        >>> model = Model(data)
        >>> model.markovModel(100, 5)
        >>> model.eqDistribution()
        """
        # logger.warning('Equilibrium distribution calculations for macrostates are now done using membership '
        #                'probabilities and hence your results might differ from analyses done before this change.')
        self._integrityCheck(postmsm=True)
        macroeq = np.ones(self.macronum) * -1
        macroindexes = list(set(self.msm.metastable_assignments))
        for i in range(self.macronum):
            # macroeq[i] = np.sum(self.msm.stationary_distribution[self.macro_ofmicro == i])
            macroeq[i] = np.sum(self.msm.metastable_memberships[:, macroindexes[i]] * self.msm.stationary_distribution)

        if plot or (save is not None):
            from matplotlib import pylab as plt
            plt.ion()
            plt.figure()
            plt.bar(np.arange(self.macronum)+0.4, macroeq)
            plt.ylabel('Equilibrium probability')
            plt.xlabel('Macrostates')
            plt.xticks(np.arange(self.macronum)+0.4, range(self.macronum))
            if save is not None:
                plt.savefig(save, dpi=300, bbox_inches='tight', pad_inches=0.2)
            if plot:
                plt.show()

        return macroeq

    def _coarseP(self):
        M = self.msm.metastable_memberships
        Pcoarse = np.linalg.inv(M.T.dot(M)).dot(M.T).dot(self.P).dot(M)
        if len(np.where(Pcoarse < 0)[0]) != 0:
            raise NameError('Cannot produce coarse P matrix. Ended up with negative probabilities. Try using less macrostates.')
        return Pcoarse

    def getStates(self, states=None, statetype='macro', wrapsel='protein', alignsel='name CA', alignmol=None, samplemode='weighted', numsamples=50, simlist=None):
        """ Get samples of MSM states in Molecule classes

        Parameters
        ----------
        states : ndarray, optional
            A list of states to visualize
        statetype : ['macro','micro','cluster'], optional
            The type of state to visualize
        wrapsel : str, optional, default='protein'
            Atom selection string to use for wrapping.
            See more `here <http://www.ks.uiuc.edu/Research/vmd/vmd-1.9.2/ug/node89.html>`__
        alignsel : str, optional, default='name CA'
            Atom selection string used for aligning all frames. Set to None to disable aligning.
            See more `here <http://www.ks.uiuc.edu/Research/vmd/vmd-1.9.2/ug/node89.html>`__
        alignmol : :class:`Molecule <moleculekit.molecule.Molecule>` object
            A reference molecule onto which to align all others
        samplemode : ['weighted','random'], optional, default='weighted'
            How to obtain the samples from the states
        numsamples : int
            Number of samples (conformations) for each state.
        simlist : numpy.ndarray of :class:`Sim <htmd.simlist.Sim>` objects
            Optionally pass a different (but matching, i.e. filtered) simlist for creating the Molecules.

        Returns
        -------
        mols : ndarray of :class:`Molecule <moleculekit.molecule.Molecule>` objects
            A list of :class:`Molecule <moleculekit.molecule.Molecule>` objects containing the samples of each state

        Examples
        --------
        >>> model = Model(data)
        >>> model.markovModel(100, 5)
        >>> mols = model.getStates()
        >>> for m in mols:
        >>>     m.view()
        """
        from htmd.projections.metric import _singleMolfile
        self._integrityCheck(postmsm=(statetype != 'cluster'))
        if simlist is None:
            simlist = self.data.simlist
        else:
            if len(simlist) != len(self.data.simlist):
                raise AttributeError('Provided simlist has different number of trajectories than the one used by the model.')

        (single, molfile) = _singleMolfile(simlist)
        if not single:
            raise NameError('Visualizer does not support yet visualization of systems with different structure files. '
                            'The simlist should be created with a single molfile (for example a filtered one)')
        if alignmol is None:
            alignmol = Molecule(molfile)
        if statetype != 'macro' and statetype != 'micro' and statetype != 'cluster':
            raise NameError("'statetype' must be either 'macro', 'micro' or ''cluster'")
        if states is None:
            if statetype == 'macro':
                states = range(self.macronum)
            elif statetype == 'micro':
                states = range(self.micronum)
            elif statetype == 'cluster':
                states = range(self.data.K)
        if len(states) == 0:
            raise NameError('No ' + statetype + ' states exist in the model')

        (tmp, relframes) = self.sampleStates(states, numsamples, statetype=statetype, samplemode=samplemode)

        from htmd.config import _config
        from htmd.parallelprogress import ParallelExecutor, delayed
        # This loop really iterates over states. sampleStates returns an array of arrays
        # Don't increase njobs because it was giving errors on some systems.
        aprun = ParallelExecutor(n_jobs=1)
        mols = aprun(total=len(relframes), desc='Getting state Molecules')\
            (delayed(_loadMols)(self, rel, molfile, wrapsel, alignsel, alignmol, simlist) for rel in relframes)
        return np.array(mols, dtype=object)

    def viewStates(self, states=None, statetype='macro', protein=None, ligand=None, viewer=None, mols=None,
                   numsamples=50, wrapsel='protein', alignsel='name CA', gui=False, simlist=None):
        """ Visualize macro/micro/cluster states in VMD

        Parameters
        ----------
        states : ndarray, optional
            A list of states to visualize
        statetype : ['macro','micro','cluster'], optional
            The type of state to visualize
        protein : bool, optional
            Set to True to enable pure protein system visualization
        ligand : str, optional
            Atom selection string for the ligand.
            See more `here <http://www.ks.uiuc.edu/Research/vmd/vmd-1.9.2/ug/node89.html>`__
        viewer : :class:`VMD <moleculekit.vmdviewer.VMD>` object, optional
            A viewer in which to visualize the states
        mols : ndarray, optional
            An array of :class:`Molecule <moleculekit.molecule.Molecule>` objects to visualize
        numsamples : int
            Number of samples (conformations) for each state.
        wrapsel : str, optional, default='protein'
            Atom selection string to use for wrapping.
            See more `here <http://www.ks.uiuc.edu/Research/vmd/vmd-1.9.2/ug/node89.html>`__
        alignsel : str, optional, default='name CA'
            Atom selection string used for aligning all frames. See to None to disable aligning.
            See more `here <http://www.ks.uiuc.edu/Research/vmd/vmd-1.9.2/ug/node89.html>`__
        simlist : numpy.ndarray of :class:`Sim <htmd.simlist.Sim>` objects
            Optionally pass a different (but matching, i.e. filtered) simlist for visualizing the states.

        Examples
        --------
        >>> model = Model(data)
        >>> model.markovModel(100, 5)
        >>> model.viewStates(protein=True)

        >>> model.viewStates(ligand='resname MOL')
        """
        from htmd.config import _config
        self._integrityCheck(postmsm=(statetype != 'cluster'))

        if protein is None and ligand is None:
            raise RuntimeError('You need to specify either the `protein` of `ligand` arguments for state visualization.')

        if _config['viewer'].lower() == 'ngl' or _config['viewer'].lower() == 'webgl':
            return self._viewStatesNGL(states, statetype, protein, ligand, mols, numsamples, gui=gui)

        if viewer is None:
            viewer = getCurrentViewer()
        if states is None:
            states = range(self.macronum)
        if isinstance(states, int):
            states = [states]
        if mols is None:
            mols = self.getStates(states, statetype, numsamples=numsamples, wrapsel=wrapsel, alignsel=alignsel, simlist=simlist)
        colors = [0, 1, 3, 4, 5, 6, 7, 9]
        for i, s in enumerate(states):
            viewer.loadMol(mols[i], name=statetype+' '+str(states[i]))
            if ligand is not None:
                viewer.rep('ligand', sel=ligand, color=colors[np.mod(i, len(colors))])
            if protein is not None:
                viewer.rep('protein')
            viewer.send('start_sscache')

    def _viewStatesNGL(self, states, statetype, protein, ligand, mols, numsamples, gui=False):
        from moleculekit.util import sequenceID
        if states is None:
            states = range(self.macronum)
        if isinstance(states, int):
            states = [states]
        if mols is None:
            mols = self.getStates(states, statetype, numsamples=min(numsamples, 15))
        colors = [0, 1, 3, 4, 5, 6, 7, 9]
        hexcolors = {0: '#0000ff', 1: '#ff0000', 2: '#333333', 3: '#ff6600', 4: '#ffff00', 5: '#4c4d00', 6: '#b2b2cc',
                     7: '#33cc33', 8: '#ffffff', 9: '#ff3399', 10: '#33ccff'}
        if protein is None and ligand is None:
            raise NameError('Please provide either the "protein" or "ligand" parameter for viewStates.')
        k = 0
        from nglview import NGLWidget, HTMDTrajectory
        view = NGLWidget(gui=gui)
        ref = mols[0].copy()
        for i, s in enumerate(states):
            if protein:
                mol = Molecule()
            if ligand:
                mol = ref.copy()
                mol.remove(ligand, _logger=False)
                mol.dropFrames(keep=0)
                mols[i].filter(ligand, _logger=False)
            mols[i].set('chain', '{}'.format(s))
            tmpcoo = mols[i].coords
            for j in range(mols[i].numFrames):
                mols[i].coords = np.atleast_3d(tmpcoo[:, :, j])
                if ligand:
                    mols[i].set('segid', sequenceID(mols[i].resid)+k)
                    k = int(mols[i].segid[-1])
                mol.append(mols[i])
            view.add_trajectory(HTMDTrajectory(mol))
            # Setting up representations
            if ligand:
                view[i].add_cartoon('protein', color='sstruc')
                view[i].add_hyperball(':{}'.format(s), color=hexcolors[np.mod(i, len(hexcolors))])
            if protein:
                view[i].add_cartoon('protein', color='residueindex')

        self._nglButtons(view, statetype, states)
        return view

    def _nglButtons(self, ngl_widget, statetype, states):
        # Adds buttons for enabling and disabling macrostate visualizations
        import ipywidgets
        from IPython.display import display

        container = []
        for s in states:
            w = ipywidgets.Checkbox(description="{} {}".format(statetype, s))
            w.value = True
            container.append(w)

        def updateReps(name):
            on = []
            for i, w in enumerate(container):
                if w.value:
                    on.append(i)
            ngl_widget.show_only(on)

        for w in container:
            w.on_trait_change(updateReps, "value")

        #container.append(ipywidgets.Checkbox(description="all"))

        hb = ipywidgets.HBox(container)
        display(hb)

    def save(self, filename):
        """ Save a :class:`Model <htmd.model.Model>` object to disk

        Parameters
        ----------
        filename : str
            Path of the file in which to save the object

        Examples
        --------
        >>> model = Model(data)
        >>> model.markovModel(100, 5)
        >>> model.save('./model.dat')
        """
        import pickle

        # Temporarily store data object and replace with dicts
        tmpdata = self.data
        if self.data.parent is not None:
            tmpparentdata = self.data.parent
            self.data.parent = self.data.parent.__dict__
        self.data = self.data.__dict__

        # Dump the dict
        f = open(filename, 'wb')
        pickle.dump(self.__dict__, f)
        f.close()

        # Restore data to classes
        self.data = tmpdata
        if self.data.parent is not None:
            self.data.parent = tmpparentdata
        

    def load(self, filename):
        """ Load a :class:`MetricData <htmd.metricdata.MetricData>` object from disk

        Parameters
        ----------
        filename : str
            Path to the saved MetricData object

        Examples
        --------
        >>> model = Model()
        >>> model.load('./model.dat')
        """
        import sys
        import pickle
        from htmd.metricdata import MetricData
        try:
            import pandas.indexes
        except ImportError:
            import pandas.core.indexes
            sys.modules['pandas.indexes'] = pandas.core.indexes  # Hacky fix for new pandas version

        f = open(filename, 'rb')
        z = pickle.load(f)
        f.close()
        for k in z:
            if k == 'data':
                m = MetricData()
                m.load(z[k])
                self.__dict__[k] = m
            else:
                self.__dict__[k] = z[k]

    def copy(self):
        """ Produces a deep copy of the object

        Returns
        -------
        data : :class:`Model` object
            A copy of the current object

        Examples
        --------
        >>> model2 = model.copy()
        """
        from copy import deepcopy
        return deepcopy(self)

    def cktest(self):
        """ Conducts a Chapman-Kolmogorow test.

        Returns
        -------

        """
        from copy import deepcopy
        from pyemma.plots import plot_cktest
        msm = deepcopy(self.msm)
        ck = msm.cktest(self.macronum)
        plot_cktest(ck)

    def createCoreSetModel(self, threshold=0.5):
        """ Given an MSM this function detects the states belonging to a core set and returns a new model consisting
        only of these states.

        Parameters
        ----------
        threshold : float
            Membership probability threshold over which microstates belong to the core of a macrostate

        Returns
        -------
        newmodel :
            A new model object
        frames : list
            A list of the frames that were kept in the new model
        """
        if (threshold >= 1) or (threshold <= 0):
            raise AttributeError('threshold argument only accepts values between (0, 1)')

        def calcCoreSet(distr, assign, threshold):
            coreset = []
            for i, md in enumerate(distr):
                microofmacro = np.where(assign == i)[0]
                prob = md[microofmacro]
                tt = threshold * (prob.max() - prob.min())
                coreset.append(microofmacro[np.where(prob > tt)[0]])
            return coreset

        def coreDtraj(data, micro_ofcluster, coreset):
            corestates = np.concatenate(coreset)
            newmapping = np.ones(corestates.max()+1, dtype=int) * -1
            newmapping[corestates] = np.arange(len(corestates))
            discretetraj = [st.copy() for st in data.St]
            frames = []
            newdiscretetraj = []
            newcounts = np.zeros(len(corestates))
            for t, st in enumerate(discretetraj):
                oldmicro = None
                newtraj = []
                tframes = []
                for f, cl in enumerate(st):
                    newmicro = None
                    micro = micro_ofcluster[cl]
                    if micro == -1:  # If we are in an dropped cluster, keep old index
                        newmicro = oldmicro
                    else:
                        for co in coreset:
                            if micro in co:
                                newmicro = micro
                                oldmicro = micro
                                break
                    if newmicro is None and oldmicro is not None:
                        newtraj.append(oldmicro)
                        tframes.append(f)
                    elif newmicro is not None:
                        newtraj.append(newmicro)
                        tframes.append(f)
                mappedtraj = newmapping[np.array(newtraj, dtype=int)]
                newdiscretetraj.append(mappedtraj)
                if len(mappedtraj):
                    newcounts[:mappedtraj.max()+1] += np.bincount(mappedtraj)
                frames.append(np.array(tframes))
            #kept = np.array([i for i, x in enumerate(newdiscretetraj) if len(x) != 0])
            return np.array(newdiscretetraj, dtype=object), len(corestates), newcounts, frames

        coreset = calcCoreSet(self.msm.metastable_distributions, self.msm.metastable_assignments, threshold)
        newdata = self.data.copy()
        newSt, newdata.K, newdata.N, frames = coreDtraj(self.data, self.micro_ofcluster, coreset)
        for i, (s, fr) in enumerate(zip(newSt, frames)):
            if len(s):
                newdata.trajectories[i].cluster[fr] = s

        logger.info('Kept {} microstates from each macrostate.'.format([len(x) for x in coreset]))

        dataobjects = [newdata]
        if newdata.parent is not None:
            dataobjects.append(newdata.parent)
        for data in dataobjects:
            dat = []
            ref = []
            simstmp = []
            cluster = []
            for i, fr in enumerate(frames):
                if len(fr):
                    dat.append(data.trajectories[i].projection[fr, :])
                    ref.append(data.trajectories[i].reference[fr, :])
                    if len(data.trajectories[i].cluster):
                        cluster.append(data.trajectories[i].cluster[fr])
                    simstmp.append(data.trajectories[i].sim)
                    
            data._loadTrajectories(dat, ref, simstmp, cluster if len(cluster) else None)

        return Model(newdata), frames

    def plotFES(self, dimX, dimY, temperature, states=False, s=10, cmap=None, fescmap=None, statescmap=None, plot=True, save=None, data=None):
        """ Plots the free energy surface on any given two dimensions. Can also plot positions of states on top.

        Parameters
        ----------
        dimX : int
            Index of projected dimension to use for the X axis.
        dimY : int
            Index of projected dimension to use for the Y axis.
        temperature : float
            Simulation temperature.
        states : bool
            If True, will plot scatter plot of microstates coloured by macro state on top of FES.
        s : float
            Marker size for states.
        cmap :
            Sets the Matplotlib colormap for both `fescmap` and `statescmap`
        fescmap :
            Matplotlib colormap for the free energy surface
        statescmap:
            Matplotlib colormap for the states
        plot : bool
            If the method should display the plot of the FES
        save : str
            Path of the file in which to save the figure
        data : :class:`MetricData` object
            Optionally you can pass a different MetricData object than the one used to build the model. For example
            if the user wants to build a model on distances but wants to plot the FES on top of RMSD values. The
            MetricData object needs to have the same simlist as the Model.

        Examples
        --------
        >>> import matplotlib as plt
        >>> model.plotFES(0, 1, 300)
        >>> model.plotFES(2, 3, 300, data=otherdata, states=True, fescmap=plt.cm.gray)
        """
        self._integrityCheck(postmsm=True)
        from matplotlib import pylab as plt
        from htmd.kinetics import Kinetics

        if data is None:
            data = self.data
            microcenters = self.data.Centers[self.cluster_ofmicro, :]
        else:
            if self.data.numFrames != data.numFrames or ~np.all([s1 == s2 for s1, s2 in zip(self.data.simlist, data.simlist)]):
                raise RuntimeError('The data argument you provided uses a different simlist than the Model.')
            microcenters = np.vstack(getStateStatistic(self, data, range(self.micronum), statetype='micro'))

        if fescmap is None:
            fescmap = plt.cm.jet
        if statescmap is None:
            statescmap = plt.cm.jet
        if cmap is not None:
            fescmap = cmap
            statescmap = cmap

        if data.description is not None:
            xlabel = data.description.description[dimX]
        else:
            xlabel = 'Dimension {}'.format(dimX)
        if data.description is not None:
            ylabel = data.description.description[dimY]
        else:
            ylabel = 'Dimension {}'.format(dimY)
        title = 'Free energy surface'

        energy = -Kinetics._kB * temperature * np.log(self.msm.stationary_distribution)
        f, ax, cf = data._contourPlot(microcenters[:, dimX], microcenters[:, dimY], energy, cmap=fescmap, xlabel=xlabel, ylabel=ylabel, title=title)
        data._setColorbar(f, cf, 'kcal/mol', scientific=False)
        if states:
            colors = statescmap(np.linspace(0, 1, self.macronum))
            for m in range(self.macronum):
                macromicro = microcenters[self.macro_ofmicro == m, :]
                _ = ax.scatter(macromicro[:, dimX], macromicro[:, dimY], s=s, c=colors[m], label='Macro {}'.format(m), edgecolors='none')
            ax.legend(prop={'size': 8})

        if save is not None:
            plt.savefig(save, dpi=300, bbox_inches='tight', pad_inches=0.2)
        if plot:
            plt.show()

    def _integrityCheck(self, postmsm=False, markov=False):
        if postmsm and self._modelid is None:
            raise NameError('You need to call markovModel before calling this command')
        if not markov and self.data._dataid == self.data._clusterid and self.data._dataid != self._clusterid:
            raise NameError('After updating the MetricData object you need to call the markovModel command anew.')
        if self.data._dataid != self.data._clusterid:
            raise NameError('After modifying the data in the MetricData object you need to recluster and reconstruct the markov model.')


def _loadMols(self, rel, molfile, wrapsel, alignsel, refmol, simlist):
    frames = self.data.rel2sim(rel, simlist=simlist)
    mol = Molecule(molfile)
    trajs = []
    frs = []
    for f in frames:
        trajs.append(f.sim.trajectory[f.piece])
        frs.append(f.frame)
    mol.read(np.array(trajs), frames=np.array(frs))
    if len(wrapsel) > 0:
        mol.wrap(wrapsel)
    if (refmol is not None) and (alignsel is not None):
        mol.align(alignsel, refmol=refmol)
    return mol


def getStateStatistic(reference, data, states, statetype='macro', weighted=False, method=np.mean, axis=0, existing=False, target=None):
    """ Calculates properties of the states.

    Calculates properties of data corresponding to states. Can calculate for example the mean distances of atoms in a
    state, or the standard deviation of the RMSD in a state.

    Parameters
    ----------
    reference : :class:`Model` object or :class:`MetricData` object
        A model containing the state definitions or a MetricData object containing cluster definitions
    data : :class:`MetricData` object
        A projection corresponding to the conformations in the states of the model. The data and the model need to share
        the same `simlist`
    states : list
        A list of the states for which we want to calculate the properties
    statetype : ['macro','micro','cluster']
        The state type
    weighted : bool
        If the properties of the macrostates should be calculated as the weighted average of their microstates, where the
        weights are the equilibrium probabilities of the microstates
    method :
        A function pointer for the method that calculates our desired property. e.g. np.mean, np.std
    axis : int
        On which axis of the data the method should operate on
    existing : bool
        Not used
    target :
        Not used

    Returns
    -------
    statistic : list
        A list which contains in each element the desired statistic of a specific state specifies in `states`

    Examples
    --------
    >>> data = MetricDistance.project(sims, 'protein and name CA', 'resname MOL')
    >>> model = Model(data)
    >>> model.markovModel(100, 5)
    >>> # Get the standard deviation of distances in all macrostates
    >>> getStateStatistic(model, data, list(range(5)), method=np.std)
    """
    from htmd.metricdata import MetricData
    if axis != 0:
        logger.warning('Axis different than 0 might not work correctly yet')

    if isinstance(reference, Model):
        refdata = reference.data
    elif isinstance(reference, MetricData):
        refdata = reference
        if statetype != 'cluster':
            raise RuntimeError('You can only use statetype cluster with reference MetricData object. To use other statetypes build a Model.')
    else:
        raise RuntimeError('Invalid argument type')

    if refdata.numTrajectories > 0 and np.any(refdata.trajLengths != data.trajLengths):
        raise NameError('Data trajectories need to match in size and number to the trajectories in the model')
    stconcat = np.concatenate(refdata.St)
    datconcat = np.concatenate(data.dat)

    statistic = []
    for i, st in enumerate(states):
        if statetype == 'macro':
            frames = reference.macro_ofcluster[stconcat] == st
        elif statetype == 'micro':
            frames = reference.micro_ofcluster[stconcat] == st
        elif statetype == 'cluster':
            frames = stconcat == st
        else:
            raise NameError('No valid state type given (read documentation)')

        if statetype == 'macro' and weighted:
            statistic.append(_weightedMethod(reference, method, stconcat, datconcat, st, axis))
        else:
            if axis is None:
                statistic.append(method(datconcat[frames, ...]))
            else:
                statistic.append(method(datconcat[frames, ...], axis=axis))
    return statistic


def _weightedMethod(model, method, stconcat, datconcat, st, axis):
    microsofmacro = np.where(model.macro_ofmicro == st)[0]
    eq = model.msm.stationary_distribution
    weights = eq / np.sum(eq[microsofmacro])
    avgstatistic = np.zeros(np.size(datconcat, 1))
    for m in microsofmacro:
        frames = model.micro_ofcluster[stconcat] == m
        if axis is None:
            stat = method(datconcat[frames, :])
        else:
            stat = method(datconcat[frames, :], axis=axis)
        avgstatistic = avgstatistic + stat * weights[m]
    return avgstatistic


def macroAccumulate(model, microvalue):
    """ Accumulate values of macrostates from a microstate array

    Parameters
    ----------
    model : :class:`Model <htmd.model.Model>` object
        The model which to use to accumulate
    microvalue : ndarray
        An array of values corresponding to the microstates of the model

    Returns
    -------
    macrovalue : ndarray
        An array of values corresponding to the macrostates of the model
    """
    res = np.zeros(model.macronum)
    for i in range(len(microvalue)):
        macro = model.macro_ofmicro[i]
        res[macro] = res[macro] + microvalue[i]
    return res


def _sampleMacro(obj, macro, stConcat, mode, numFrames, replacement):
    from htmd.metricdata import _randomSample
    if mode == 'random':
        frames = np.where(obj.macro_ofcluster[stConcat] == macro)[0]
        selFrames = _randomSample(frames, numFrames, replacement)
        selMicro = obj.micro_ofcluster[stConcat[selFrames]]
    elif mode == 'even':
        micros = np.where(obj.macro_ofmicro == macro)[0]
        selFrames = []
        selMicro = []
        for i in range(len(micros)):
            selFrames.append(_sampleMicro(obj, micros[i], stConcat, numFrames, replacement))
            selMicro.append(np.ones(numFrames) * micros[i])
    elif mode == 'weighted':
        micros = np.where(obj.macro_ofmicro == macro)[0]
        eq = obj.msm.stationary_distribution
        weights = eq[micros] / np.sum(eq[micros])
        framespermicro = np.random.multinomial(numFrames, weights)
        selFrames = []
        selMicro = []
        for i in range(len(micros)):
            selFrames = np.append(selFrames, _sampleMicro(obj, micros[i], stConcat, framespermicro[i], replacement))
            selMicro = np.append(selMicro, [micros[i]] * framespermicro[i])
    elif mode == 'weightedTrunc':
        micros = np.where(obj.macro_ofmicro == macro)[0]
        eq = obj.msm.stationary_distribution
        weights = eq[micros] / np.sum(eq[micros])
        idx = np.argsort(weights)
        cs = np.cumsum(weights[idx])
        under50 = cs < 0.5
        weights[idx[under50]] = 0
        weights /= np.sum(weights)
        framespermicro = np.random.multinomial(numFrames, weights)
        selFrames = []
        selMicro = []
        for i in range(len(micros)):
            selFrames = np.append(selFrames, _sampleMicro(obj, micros[i], stConcat, framespermicro[i], replacement))
            selMicro = np.append(selMicro, [micros[i]] * framespermicro[i])
    else:
        raise NameError('No valid mode given (read documentation)')
    return selFrames, selMicro


def _sampleMicro(obj, micro, stConcat, numFrames, replacement):
    from htmd.metricdata import _randomSample
    frames = np.where(obj.micro_ofcluster[stConcat] == micro)[0]
    return _randomSample(frames, numFrames, replacement)


def _macroTrajectoriesReport(macronum, macrost, simlist=None):
    """
    Prints out the number of trajectories that visited each macrostate and prints warnings

    Parameters
    ----------
    macronum
    macrost
    simlist
    """
    macrotraj = dict()
    macrotrajnum = np.zeros(macronum, dtype=int)
    for i, st in enumerate(macrost):
        macros = np.setdiff1d(np.unique(st), [-1])
        for m in macros:
            if m not in macrotraj:
                macrotraj[m] = []
            macrotraj[m].append(i)
            macrotrajnum[m] += 1
    logger.info('Number of trajectories that visited each macrostate:')
    logger.info(macrotrajnum)

    for m in range(macronum):
        ratio = macrotrajnum[m] / len(macrost)
        if macrotrajnum[m] <= 3 and ratio <= 0.2:
            logger.info('Take care! Macro {} has been visited only in {} trajectories'
                        ' ({:.1f}% of total):'.format(m, macrotrajnum[m], ratio*100))
            if simlist is not None:
                for s in macrotraj[m]:
                    logger.info(simlist[s])


def _macroTrajSt(St, macro_ofcluster):
    mst = np.empty(np.shape(St), dtype=object)
    for i in range(len(St)):
        mst[i] = macro_ofcluster[St[i]]
    return mst


'''def _macroP(C, macro_ofmicro):
    macronum = np.max(macro_ofmicro) + 1
    macroC = np.zeros((macronum, macronum))

    for m1 in range(macronum):
        sourcemicros = np.where(macro_ofmicro == m1)[0]
        for m2 in range(macronum):
            sinkmicros = np.where(macro_ofmicro == m2)[0]
            ixgrid = np.ix_(sourcemicros, sinkmicros)
            macroC[m1, m2] = np.sum(C[ixgrid].flatten())

    from msmtools.estimation import transition_matrix
    return transition_matrix(macroC, reversible=True)'''

import unittest
class _TestModel(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        from htmd.simlist import simlist, simfilter
        from glob import glob
        from htmd.projections.metric import Metric
        from moleculekit.projections.metricdistance import MetricDistance
        from moleculekit.projections.metricdihedral import MetricDihedral
        from moleculekit.util import tempname
        from sklearn.cluster import MiniBatchKMeans
        from htmd.home import home
        from os.path import join

        sims = simlist(glob(join(home(dataDir='adaptive'), 'data', '*', '')), glob(join(home(dataDir='adaptive'), 'input', '*')))
        fsims = simfilter(sims, tempname(), 'not water')

        metr = Metric(fsims)
        metr.set(MetricDistance('protein and resid 10 and name CA', 'resname BEN and noh', metric='contacts', groupsel1='residue', threshold=4))
        data = metr.project()
        data.cluster(MiniBatchKMeans(n_clusters=4))

        self.model = Model(data)

    def test_model_saving_loading(self):
        from moleculekit.util import tempname

        modelfile = tempname(suffix='.dat')
        self.model.save(modelfile)

        newmodel = Model(file=modelfile)
        assert newmodel.data.numTrajectories == 2

        # Testing model saving when data has parent
        self.model.data.parent = self.model.data.copy()
        self.model.save(modelfile)
        newmodel = Model(file=modelfile)
        assert newmodel.data.numTrajectories == 2
        assert newmodel.data.parent.numTrajectories == 2
        self.model.data.parent = None

if __name__ == '__main__':
    unittest.main(verbosity=2)






# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
import numpy as np
from htmd.kinetics import Rates
import logging
logger = logging.getLogger(__name__)


class KineticsHMM(object):
    _kB = 0.0019872041  # Boltzman constant kcal/(mol K)

    def __init__(self, model, temperature, concentration=1, source=None, sink=None):
        self.model = model
        self.concentration = concentration
        self.temperature = temperature
        self.source = source
        self.sink = sink

        # if model._modelid is None:
        #     raise RuntimeError('You need to call model.markovModel() before calculating kinetics')
        #
        # self._modelid = model._modelid

        if self.source is None:
            logger.info('Detecting source state...')
            self._detectSource()
            logger.info('Source macro = ' + str(model.hmm.active_set[self.source]))
        if self.sink is None:
            logger.info('Detecting sink state...')
            self._detectSink()
            logger.info('Sink macro = ' + str(model.hmm.active_set[self.sink]))
        if model.data.fstep is None:
            raise RuntimeError('Please define a framestep (fstep) for the data object')

    def _detectSource(self):
        if self.model.data.parent is None:
            dataobj = self.model.data
        else:
            dataobj = self.model.data.parent
        St = self.model.hmm.discrete_trajectories_full
        K = self.model.data.K

        distcols = []
        contcols = []
        if not dataobj.description.empty:  # Search for distances or contacts in the data
            distcols = np.where(dataobj.description.type == 'distance')[0]
            contcols = np.where(dataobj.description.type == 'contact')[0]
        if len(distcols) == 0 and len(contcols) == 0:
            raise RuntimeError('Could not detect source state. Please specify it in the Kinetics constructor.')

        if len(distcols) != 0:
            cols = distcols
        else:
            cols = contcols

        averages = np.zeros(self.model.hmm.nstates)
        N = np.zeros(self.model.hmm.nstates)
        dat = dataobj.dat
        for i in range(len(St)):
            if dataobj.trajectories[i].projection.ndim == 2:
                rowsums = np.sum(dat[i][:, cols], axis=1)
            else:
                rowsums = dat[i][:, cols]
            macroSt = self.model.hmm.metastable_assignments[St[i]]
            totalsums = np.bincount(macroSt, weights=rowsums)
            N[0:np.max(macroSt)+1] += np.bincount(macroSt)
            idx = list(set(macroSt))
            averages[idx] += totalsums[idx]
        avg = averages / N
        avg = avg[self.model.hmm.active_set]

        if len(contcols) != 0:
            logger.info('Guessing the source state as the state with minimum contacts.')
            self.source = np.argmin(avg)
        elif len(distcols) != 0:
            logger.info('Guessing the source state as the state with maximum distances.')
            self.source = np.argmax(avg)

    def _detectSink(self):
        if self.source is None:
            raise NameError('Source undefined. Cannot define sink.')
        peq = self.model.eqDistribution(plot=False)
        idx = np.argsort(peq)
        idx = np.delete(idx, np.where(idx == self.source))
        self.sink = idx[-1]

    def getRates(self, source=None, sink=None):
        """ Get the rates between two states

        Parameters
        ----------
        source : int, optional
            The state index to use as source
        sink : int, optional
            The state index to use as sink

        Returns
        -------
        rates : :class:`Rates` object
            A Rates object containing the rates

        Example
        -------
        >>> kin = Kinetics(model, temperature=300, concentration=0.015)
        >>> r = kin.getRates()
        >>> print(r)
        >>> dg = r.g0eq
        """
        actset = self.model.hmm.active_set
        if source is None:
            source = self.source
        else:
            source = np.where(actset == source)[0]

        if sink is None:
            sink = self.sink
        else:
            sink = np.where(actset == sink)[0]
        logger.info('Calculating rates between source: {} and sink: {} states.'.format(actset[source], actset[sink]))
        if source == sink:
            logger.info('Calculating rates between state and itself gives 0')
            r = Rates(); r.mfpton = 0; r.mfptoff=0; r.koff=0; r.kon=0; r.g0eq=0; r.kdeq=0;
            return r

        if source == self.source:  # Apply concentration only on the bulk state
            conc = self.concentration
        elif sink == self.source:
            conc = 1 / self.concentration
        else:
            conc = 1

        model = self.model
        from msmtools.analysis import mfpt
        r = Rates()
        r.mfpton = model.data.fstep * model.hmm.lag * mfpt(self.model.hmm.transition_matrix, origin=source, target=sink)
        r.mfptoff = model.data.fstep * model.hmm.lag * mfpt(self.model.hmm.transition_matrix, origin=sink, target=source)
        r.koff = 1E9 / r.mfptoff
        r.kon = 1E9 / (r.mfpton * conc)
        eq = model.hmm.stationary_distribution
        sinkeq = np.sum(eq[sink])
        sourceeq = np.sum(eq[source])
        if conc != 1:
            logger.info('Concentration correction of {:.2f} kcal/mol.'.format(-self._kBT * np.log(1 / conc)))
        r.g0eq = -self._kBT * np.log(sinkeq / (conc * sourceeq))
        r.kdeq = np.exp(r.g0eq / self._kBT)
        return r

    @property
    def _kBT(self):
        return self._kB * self.temperature






# (c) 2015-2019 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
import json
import logging
import os
import shutil
import sys
import tempfile
import time
import zipfile

from jobqueues.simqueue import SimQueue
from protocolinterface import ProtocolInterface
from protocolinterface.validators import Number, String

logger = logging.getLogger(__name__)

class PlayQueue(SimQueue, ProtocolInterface):

    def __init__(self):

        SimQueue.__init__(self)
        ProtocolInterface.__init__(self)

        self._arg('ngpu', 'int', 'Number of GPUs', default=0, validator=Number(int, '0POS'))
        self._arg('ncpu', 'int', 'Number of CPUs', default=1, validator=Number(int, '0POS'))
        self._arg('memory', 'int', 'Amount of memory (MB)', default=1000, validator=Number(int, 'POS'))
        self._arg('max_jobs', 'int', 'Maximum number of concurent jobs', default=sys.maxsize, validator=Number(int, 'POS'))

        self._arg('token', 'str', 'PM token', required=True, validator=String())
        self._arg('app', 'str', 'App name', required=True, validator=String())

        self._dirs = {}

    def _getSession(self):

        from playmolecule import Session

        return Session(self.token)

    @staticmethod
    def _getCurrentJobID():
        # TODO this could be part of PM API

        configFile = '/data/in/config'

        if not os.path.exists(configFile):
            return None

        with open(configFile) as fd:
            return json.load(fd)['execid']

    def _makeZIP(self, dir_):

        zipFile = os.path.join(dir_, 'input.zip')

        files = os.listdir(dir_)
        with zipfile.ZipFile(zipFile, 'w') as zf:
            for file in files:
                zf.write(os.path.join(dir_, file), arcname=file)

        return zipFile

    def submit(self, dirs):

        logger.info(f'Maximum number of concurrent jobs: {self.max_jobs}')

        dirs = [dirs, ] if isinstance(dirs, str) else dirs
        for dir_ in dirs:

            # Delay submission
            while self.inprogress() >= self.max_jobs:
                time.sleep(5)

            # Submit a job
            job = self._getSession().startApp(self.app)
            job.input = self._makeZIP(dir_) # TODO this is Psi4 specific
            job.submit(_logger=False, childOf=self._getCurrentJobID())
            self._dirs[job._execid] = dir_
            logger.info(f'Submitted job {job._execid}:')
            logger.info(f'    App name: {self.app}')
            logger.info(f'    Input directory: {dir_}')
            logger.info(f'    Resources: {self.ngpu} GPUs, {self.ncpu} CPUs, {self.memory} MB of memory')

    def inprogress(self):

        # TODO updated to use playmolecule.Job.getChildren
        #      see https://github.com/Acellera/htmd/pull/910

        counter = 0
        for jobID in self._dirs:
            job = self._getSession().getJob(id=jobID)
            status = job.getStatus(_logger=False)
            if status in (0, 1, 2, 3, 6, 7): # Queuing, running, etc.
                counter += 1
            elif status in (4, 5): # Completed or errored
                pass
            else:
                raise ValueError('Unknown job status')

        return counter

    def retrieve(self):

        for jobID, dir_ in list(self._dirs.items()):
            job = self._getSession().getJob(id=jobID)
            with tempfile.TemporaryDirectory() as tmpDir:
                status = job.getStatus(_logger=False)
                if status == 4:
                    logger.info(f'Job {jobID} completed (status: {status}):')
                    outDir = job.retrieve(_logger=False, path=tmpDir)
                    for file in os.listdir(outDir):
                        shutil.copy(os.path.join(outDir, file), dir_)
                    self._dirs.pop(jobID)
                    logger.info(f'    Retrieved results to {dir_}')
                elif status == 5:
                    logger.info(f'Job {jobID} failed (status: {status})')
                    self._dirs.pop(jobID)
                elif status in (0, 1, 2, 3, 6, 7):
                    pass
                else:
                    raise ValueError('Unknown job status')

    def stop(self):
        raise NotImplementedError()

    @property
    def ncpu(self):
        return self.__dict__['ncpu']

    @ncpu.setter
    def ncpu(self, value):
        self.ncpu = value

    @property
    def ngpu(self):
        return self.__dict__['ngpu']

    @ngpu.setter
    def ngpu(self, value):
        self.ngpu = value

    @property
    def memory(self):
        return self.__dict__['memory']

    @memory.setter
    def memory(self, value):
        self.memory = value



from moleculekit.projections.metricsecondarystructure import MetricSecondaryStructure
import logging


logger = logging.getLogger(__name__)

logger.warning('Please do not import from htmd.projections.metricsecondarystructure. This will be deprecated. '
                'Use moleculekit.projections.metricsecondarystructure instead.')


from moleculekit.projections.metricshell import MetricShell
import logging


logger = logging.getLogger(__name__)

logger.warning('Please do not import from htmd.projections.metricshell. This will be deprecated. '
                'Use moleculekit.projections.metricshell instead.')


from jobqueues.simqueue import SimQueue, RetrieveError, SubmitError, InProgressError, ProjectNotExistError
import logging


logger = logging.getLogger(__name__)

logger.warning('Please do not import from htmd.queues.simqueue. This will be deprecated. '
                'Use jobqueues.simqueue instead.')



# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
from moleculekit.molecule import Molecule
from htmd.apps.acemd import Acemd
from htmd.protocols.oldprotocolinterface import ProtocolInterface, TYPE_INT, TYPE_FLOAT, RANGE_0POS, RANGE_POS, RANGE_ANY
import os
import numpy as np
import logging
from htmd.decorators import _Deprecated
logger = logging.getLogger(__name__)


@_Deprecated('1.13.6')
class Equilibration(ProtocolInterface):
    """ Equilibration protocol

        Equilibration protocol for globular and membrane proteins
        It includes a flatbottom potential box to retrain a ligand
        for example within this box.

        Parameters
        ----------
        numsteps : int, default=0
            Number of steps to run the simulations in units of 4fs
        temperature : float, default=300
            Temperature of the thermostat in Kelvin
        k : float, default=0
            Force constant of the flatbottom potential in kcal/mol/A^2. E.g. 5
        reference : str, default='none'
            Reference selection to use as dynamic center of the flatbottom box.
        selection : str, default='none'
            Selection of atoms to apply the flatbottom potential
        box : list, default=[0, 0, 0, 0, 0, 0]
            Position of the flatbottom box in term of the reference center given as [xmin, xmax, ymin, ymax, zmin, zmax]
        useconstantratio : bool, default=False
            For membrane protein simulations set it to true so that the barostat does not modify the xy aspect ratio.
        constraints : dict, default={'protein and noh and not name CA': 0.1, 'protein and name CA': 1}
            A dictionary containing as keys the atomselections of the constraints and as values the constraint scaling factor. 0 factor means no constraint, 1 full constraints and in between values are used for scaling. The order with which the constraints are applied is random, so make atomselects mutually exclusive to be sure you get the correct constraints.

        Example
        -------
        >>> from htmd.protocols.equilibration_v1 import Equilibration
        >>> md = Equilibration()
        >>> md.numsteps = 10000000
        >>> md.temperature = 300
        >>> md.useconstantratio = True  # only for membrane sims
        >>> # this is only needed for setting the flatbottom potential, otherwise remove it
        >>> md.reference = 'protein and resid 293'
        >>> md.selection = 'segname L and noh'
        >>> md.box = [-25, 25, -25, 25, 43, 45]
        >>> md.k = 5
        >>> md.write('./build','./equil')
    """
    def __init__(self):
        super().__init__()
        self._cmdObject('acemd', ':class:`MDEngine <htmd.apps.app.App>` object', 'MD engine', None, Acemd)
        self._cmdValue('numsteps', 'int', 'Number of steps to run the simulations in units of 4fs', 0, TYPE_INT, RANGE_0POS)
        self._cmdValue('temperature', 'float', 'Temperature of the thermostat in Kelvin', 300, TYPE_FLOAT, RANGE_ANY)
        self._cmdValue('k', 'float', 'Force constant of the flatbottom potential in kcal/mol/A^2. E.g. 5', 0, TYPE_FLOAT, RANGE_ANY)
        self._cmdString('reference', 'str', 'Reference selection to use as dynamic center of the flatbottom box.', 'none')
        self._cmdString('selection', 'str', 'Selection of atoms to apply the flatbottom potential', 'none')
        self._cmdList('box', 'list', 'Position of the flatbottom box in term of the reference center given as [xmin, xmax, ymin, ymax, zmin, zmax]', [0,0,0,0,0,0])
        self._cmdBoolean('useconstantratio', 'bool', 'For membrane protein simulations set it to true so that the barostat does not modify the xy aspect ratio.', False)
        self._cmdDict('constraints', 'dict', 'A dictionary containing as keys the atomselections of the constraints '
                                             'and as values the constraint scaling factor. 0 factor means no constraint'
                                             ', 1 full constraints and in between values are used for scaling.'
                                             ' The order with which the constraints are applied is random, so make '
                                             'atomselects mutually exclusive to be sure you get the correct constraints.'
                                             , {'protein and noh and not name CA': 0.1, 'protein and name CA': 1})

        self.acemd = Acemd()
        self.acemd.coordinates = None
        self.acemd.structure = None
        self.acemd.parameters = None
        self.acemd.temperature = '$temperature'
        self.acemd.restart = 'on'
        self.acemd.restartfreq = '5000'
        self.acemd.outputname = 'output'
        self.acemd.xtcfile = 'output.xtc'
        self.acemd.xtcfreq = '25000'
        self.acemd.timestep = '4'
        self.acemd.rigidbonds = 'all'
        self.acemd.hydrogenscale = '4'
        self.acemd.switching = 'on'
        self.acemd.switchdist = '7.5'
        self.acemd.cutoff = '9'
        self.acemd.exclude = 'scaled1-4'
        self.acemd.scaling14 = '1.0'
        self.acemd.langevin = 'on'
        self.acemd.langevintemp = '$temperature'
        self.acemd.langevindamping = '1'
        self.acemd.pme = 'on'
        self.acemd.pmegridspacing = '1.0'
        self.acemd.fullelectfrequency = '2'
        self.acemd.energyfreq = '1000'
        self.acemd.constraints = 'on'
        self.acemd.consref = None
        self.acemd.constraintscaling = '1.0'
        self.acemd.berendsenpressure = 'on'
        self.acemd.berendsenpressuretarget = '1.01325'
        self.acemd.berendsenpressurerelaxationtime = '800'
        self.acemd.tclforces = 'on'
        self.acemd.minimize = '500'
        self.acemd.run = '$numsteps'
        self.acemd.TCL='''
set numsteps NUMSTEPS
set temperature TEMPERATURE
set refindex { REFINDEX }
set selindex { SELINDEX }
set box { BOX }
set K KCONST
#
proc flatbot1d {x xm xM K} {
  set f 0
  if {$x < $xm} {
    set f [expr $K*[expr $xm-$x]]
  }
  if {$x > $xM} {
    set f [expr $K*[expr $xM-$x]]
  }
  return $f
}
proc calcforces_init {} {
  global ref sel refindex selindex
  berendsenpressure  off
  set ref [addgroup  $refindex]
  set sel [addgroup  $selindex]
}
proc calcforces {} {
  global ref sel numsteps K box
  loadcoords coords
##FLATBOTTOM
  if {$K>0} {
    set r0 $coords($ref)
    set r1 $coords($sel)
    set dr  [vecsub $r1 $r0]
    set fx [flatbot1d [lindex $dr 0] [lindex $box 0] [lindex $box 1] $K]
    set fy [flatbot1d [lindex $dr 1] [lindex $box 2] [lindex $box 3] $K]
    set fz [flatbot1d [lindex $dr 2] [lindex $box 4] [lindex $box 5] $K]
    #print "dr: $dr  fx: $fx fy: $fy fz: $fz"
    addforce $sel [list $fx $fy $fz]
  }
##EQUIL
  set step [ getstep ]
  if { $step > 500 } {
    berendsenpressure  on
  } else {
    berendsenpressure  off}
  if { $step > [expr $numsteps/2] } {
    constraintscaling 0
  } else {
    constraintscaling [expr 1 + $step*(0.05-1)*2/$numsteps]}
}
proc calcforces_endstep { } { }
proc calcforces_terminate { } { }
'''

    def _findFiles(self, inputdir):
        # Tries to find default files if the given don't exist
        defaults = {'coordinates': ('structure.pdb', ),
                    'structure': ('structure.psf', 'structure.prmtop'),
                    'parameters': ('parameters', 'structure.prmtop')}

        for field in defaults:
            userval = self.acemd.__dict__[field]
            if userval is not None and not os.path.exists(os.path.join(inputdir, userval)):
                self.acemd.__dict__[field] = None

            if self.acemd.__dict__[field] is None:
                for val in defaults[field]:
                    if os.path.exists(os.path.join(inputdir, val)):
                        self.acemd.__dict__[field] = val
                        break

            if userval is not None and self.acemd.__dict__[field] is not None and self.acemd.__dict__[field] != userval:
                logger.warning('Could not locate structure file {}. Using {} instead.'.format(
                    os.path.join(inputdir, userval), os.path.join(inputdir, self.acemd.__dict__[field])
                ))
            elif self.acemd.__dict__[field] is None:
                raise RuntimeError('Could not locate any {f:} file in {i:}. '
                                   'Please set the Equilibration.acemd.{f:} property to '
                                   'point to the {f:} file'.format(f=field, i=inputdir))

        if self.acemd.consref is None:
            self.acemd.consref = self.acemd.coordinates

    def _amberFixes(self):
        # AMBER specific fixes
        if self.acemd.parameters.endswith('structure.prmtop'):
            self.acemd.parmfile = self.acemd.parameters
            self.acemd.parameters = None
            self.acemd.scaling14 = '0.8333333'
            self.acemd.amber = 'on'

    def write(self, inputdir, outputdir):
        """ Write the equilibration protocol

        Writes the equilibration protocol and files into a folder for execution
        using files inside the inputdir directory

        Parameters
        ----------
        inputdir : str
            Path to a directory containing the files produced by a build process.
        outputdir : str
            Directory where to write the equilibration setup files.

        Examples
        --------
        >>> md = Equilibration()
        >>> md.write('./build','./equil')
        """
        self._findFiles(inputdir)
        self._amberFixes()

        pdbfile = os.path.join(inputdir, self.acemd.coordinates)
        inmol = Molecule(pdbfile)

        self.acemd.TCL = self.acemd.TCL.replace('NUMSTEPS', str(self.numsteps))
        self.acemd.TCL = self.acemd.TCL.replace('TEMPERATURE', str(self.temperature))
        self.acemd.TCL = self.acemd.TCL.replace('KCONST', str(self.k))
        self.acemd.TCL = self.acemd.TCL.replace('REFINDEX', ' '.join(map(str, inmol.get('index', self.reference))))
        self.acemd.TCL = self.acemd.TCL.replace('SELINDEX', ' '.join(map(str, inmol.get('index', self.selection))))
        self.acemd.TCL = self.acemd.TCL.replace('BOX', ' '.join(map(str, self.box)))
        if self.acemd.celldimension is None and self.acemd.extendedsystem is None:
            coords = inmol.get('coords', sel='water')
            if coords.size == 0:  # It's a vacuum simulation
                coords = inmol.get('coords', sel='all')
                dim = np.max(coords, axis=0) - np.min(coords, axis=0)
                dim = dim + 12.
            else:
                dim = np.max(coords, axis=0) - np.min(coords, axis=0)
            self.acemd.celldimension = '{} {} {}'.format(dim[0], dim[1], dim[2])
        if self.useconstantratio:
            self.acemd.useconstantratio = 'on'
        self.acemd.setup(inputdir, outputdir, overwrite=True)

        # Adding constraints
        inmol.set('occupancy', 0)
        inmol.set('beta', 0)
        for sel in self.constraints:
            inmol.set('beta', self.constraints[sel], sel)
        outfile = os.path.join(outputdir, self.acemd.coordinates)
        inmol.write(outfile)

    def addConstraint(self, atomselect, factor=1):
        """ Convenience function for adding a new constraint to existing constraints.

        Parameters
        ----------
        atomselect : str
            Atom selection of atoms we want to constrain
        factor : float
            The scaling factor of the constraints applied to the atoms

        Example
        -------
        >>> eq.addConstraint('chain X', 0.3)
        """
        self.constraints[atomselect] = factor

if __name__ == "__main__":
    import htmd.home
    eq = Equilibration()
    eq.numsteps = 1000000
    eq.temperature = 300
    eq.reference = 'protein and name CA'
    eq.selection = 'segname L and noh'
    eq.box = [-20, 20, -20, 20, 43, 45]
    eq.k = 5
    eq.write(htmd.home.home() + '/data/equilibrate', '/tmp/equil1')



from moleculekit.projections.metricdihedral import *
import logging


logger = logging.getLogger(__name__)

logger.warning('Please do not import from htmd.projections.metricdihedral. This will be deprecated. '
                'Use moleculekit.projections.metricdihedral instead.')


# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
import setuptools
import subprocess

try:
    version = (
        subprocess.check_output(["git", "describe", "--abbrev=0", "--tags"])
        .strip()
        .decode("utf-8")
    )
except Exception as e:
    print("Could not get version tag. Defaulting to version 0")
    version = "0"

with open("requirements.txt", "r") as f:
    requirements = f.read().splitlines()


if __name__ == "__main__":
    with open("README.md", "r") as fh:
        long_description = fh.read()

    setuptools.setup(
        name="acellera-htmd",
        version=version,
        description="High throughput molecular dynamics (HTMD)",
        long_description=long_description,
        zip_safe=False,
        url="https://www.htmd.org",
        maintainer="Acellera Ltd",
        maintainer_email="info@acellera.com",
        entry_points={
            "console_scripts": [
                "htmd_register   = htmdx.cli:htmd_do_register",
                "activate_license  = htmdx.cli:main_activate",
            ]
        },
        packages=setuptools.find_packages(include=["htmd*"], exclude=["data"]),
        package_data={"htmd": ["share/*/*/*/*"],},
        install_requires=requirements,
    )



from moleculekit.projections.metricsphericalcoordinate import MetricSphericalCoordinate
import logging


logger = logging.getLogger(__name__)

logger.warning('Please do not import from htmd.projections.metricsphericalcoordinate. This will be deprecated. '
                'Use moleculekit.projections.metricsphericalcoordinate instead.')


# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
import abc
from difflib import get_close_matches
import os
import re
from htmd.util import ensurelist

RANGE_ANY = 0
RANGE_POS = 1
RANGE_NEG = 2
RANGE_0POS = 3
RANGE_0NEG = 4

TYPE_INT = 0
TYPE_FLOAT = 1


class ProtocolInterface:
    def __init__(self):
        self._commands = {}
        self._deprecated = {}
        self._numCommands = 0
        self._commandsOrder = {}

    def __setattr__(self, key, value):
        if key[0] != '_':
            self._validate(key, value)
        self.__dict__[key] = value

    def __str__(self):
        sortedkeys = [x[0] for x in sorted(self._commandsOrder.items(), key=lambda x: x[1])]
        s = ''
        for cmd in sortedkeys:
            if isinstance(self.__dict__[cmd], str):
                s += '{} = \'{}\'\n'.format(cmd, self.__dict__[cmd])
            else:
                s += '{} = {}\n'.format(cmd, self.__dict__[cmd])
        return s

    def __misc(self, key, default):
        self._commandsOrder[key] = self._numCommands
        self._numCommands += 1
        self.__dict__[key] = default

    def _cmdFile(self, key, datatype, descr, default, exist=False, writable=False, multiple=False, check=None):
        self._commands[key] = FileValidator(key, datatype, descr, default, exist, writable, multiple, check)
        self.__misc(key, default)

    def _cmdListList(self, key, datatype, descr, length, default=None):
        self._commands[key] = ListListValidator(key, datatype, descr, default, length)
        self.__misc(key, default)

    def _cmdString(self, key, datatype, descr, default, valid_values=None):
        self._commands[key] = StringValidator(key, datatype, descr, default, valid_values)
        self.__misc(key, default)

    def _cmdTimestep(self, key, datatype, descr, default):
        self._commands[key] = TimestepValidator(key, datatype, descr, default)
        self.__misc(key, default)

    def _cmdList(self, key, datatype, descr, default, valid_values=None):
        self._commands[key] = ListValidator(key, datatype, descr, default, valid_values)
        self.__misc(key, default)

    def _cmdDict(self, key, datatype, descr, default, key_type=None, val_type=None):
        self._commands[key] = DictionaryValidator(key, datatype, descr, default, key_type, val_type)
        self.__misc(key, default)

    def _cmdBinary(self, key, datatype, descr, default):
        self._commands[key] = BinaryValidator(key, datatype, descr, default)
        self.__misc(key, default)

    def _cmdValue(self, key, datatype, descr, default, realdatatype, valid_range, list_len=1, units=None):
        self._commands[key] = ValueValidator(key, datatype, descr, default, realdatatype, valid_range, list_len, units)
        self.__misc(key, default)

    def _cmdObject(self, key, datatype, descr, default, classname):
        self._commands[key] = ObjectValidator(key, datatype, descr, default, classname)
        self.__misc(key, default)

    def _cmdClass(self, key, datatype, descr, default, classname):
        self._commands[key] = ClassValidator(key, datatype, descr, default, classname)
        self.__misc(key, default)

    def _cmdBoolean(self, key, datatype, descr, default):
        self._commands[key] = BooleanValidator(key, datatype, descr, default)
        self.__misc(key, default)

    def _cmdFunction(self, key, datatype, descr, default):
        self._commands[key] = FunctionValidator(key, datatype, descr, default)
        self.__misc(key, default)

    def _cmdDeprecated(self, key, newkey=None):
        self._deprecated[key] = newkey

    def _printDocString(self):
        sortedkeys = [x[0] for x in sorted(self._commandsOrder.items(), key=lambda x: x[1])]
        docs = ''
        for k in sortedkeys:
            docs += '{} : {}\n'.format(k, self._commands[k].docString())
        print(docs)

    def _validate(self, key, value, basedir=None):
        if key in self._deprecated:
            newkey = self._deprecated[key]
            if not newkey:
                print(" Attribute '" + key + "'\t is deprecated and is no longer required")
            else:
                print(" Attribute '" + key + "'\t is deprecated and replaced by '" + newkey + "'")
                key = newkey

        if key not in self._commands:
            strerror = "Attribute '" + key + "' not allowed in this class."
            match = get_close_matches(key, self._commands)
            if match:
                strerror += " Try '" + match[0] + "'"
            raise ValueError(strerror)

        cmd = self._commands[key]
        return cmd.validate(value, basedir=basedir)

    def _toArgParse(self, description, parser=None):
        import argparse
        if parser is None:
            parser = argparse.ArgumentParser(description=description)

        sortedkeys = [x[0] for x in sorted(self._commandsOrder.items(), key=lambda x: x[1])]
        for k in sortedkeys:
            parser.add_argument('--{}'.format(k), help=self._commands[k].descr, default=self._commands[k].default)

        return parser

    def _fromArgParse(self, args):
        argsdict = vars(args)
        for k in argsdict:
            self.__dict__[k] = argsdict[k]


class Validator:
    def __init__(self, key, datatype, descr, default):
        self.datatype = datatype
        if not descr:
            desc = os.path.join("help", "en", key)
        if descr and os.path.exists(descr):
            f = open(descr, 'r')
            self.descr = f.read()
            f.close()
        else:
            self.descr = descr
        self.default = default

    def docString(self):
        doc = ''
        if hasattr(self, 'valid_values') and self.valid_values is not None:
            doc += '{}, '.format(self.valid_values)
        doc += '{}'.format(self.datatype)
        # if self.default is not None:
        if isinstance(self.default, str):
            doc += ', default=\'{}\''.format(self.default)
        else:
            doc += ', default={}'.format(self.default)
        if hasattr(self, 'units') and self.units is not None:
            doc += ', units={}'.format(self.units)
        doc += '\n\t{}'.format(self.descr)
        return doc


# TODO: Maybe better to consolidate it into ValueValidator
class BooleanValidator(Validator):
    def __init__(self, key, datatype, descr, default):
        super().__init__(key, datatype, descr, default)

    def args(self):
        return

    def validate(self, object, basedir=None):
        return isinstance(object, bool)


class ObjectValidator(Validator):
    def __init__(self, key, datatype, descr, default, classname):
        super().__init__(key, datatype, descr, default)
        self.classname = classname

    def args(self):
        return

    def validate(self, object, basedir=None):
        classname = self.classname
        object = ensurelist(object)
        classname = ensurelist(classname)

        for obj in object:
            valid = False
            for cl in classname:
                if isinstance(obj, cl):
                    valid = True
                    break
            if not valid:
                raise ValueError('Value must be object of {}'.format(self.classname))

        return object

class ClassValidator(Validator):
    def __init__(self, key, datatype, descr, default, classname):
        super().__init__(key, datatype, descr, default)
        self.classname = classname

    def args(self):
        return

    def validate(self, object, basedir=None):
        classname = self.classname
        object = ensurelist(object)
        classname = ensurelist(classname)

        for obj in object:
            valid = False
            for cl in classname:
                if issubclass(obj, cl):
                    valid = True
                    break
            if not valid:
                raise ValueError('Value must be subclass of {}'.format(self.classname))

        return object


class FunctionValidator(Validator):
    def __init__(self, key, datatype, descr, default):
        super().__init__(key, datatype, descr, default)

    def args(self):
        return

    def validate(self, object, basedir=None):
        if hasattr(object, '__call__'):
            return
        elif isinstance(object, tuple) and hasattr(object[0], '__call__'):
            # Supporting delayed functions which are passed as function/arguments tuples
            return
        else:
            raise ValueError('Value must be a function')


class ListListValidator(Validator):
    def __init__(self, key, datatype, descr, default, length):
        super().__init__(key, datatype, descr, default)
        self.length = length

    def args(self):
        return "A list of lists with length " + self.length

    def validate(self, value_list, basedir=None):
        vv = ""
        for t in value_list:
            vv = vv + " " + t
        vv = vv.strip()
        if vv == "[]":
            self.value = None
            return self.value

        vv = re.sub("\[", "", vv)
        vv = re.sub(",", "", vv)
        vv = re.sub("\]\]$", "", vv)
        vv = re.sub("\],", ",", vv)
        vv = re.sub("'", "", vv)
        vv = vv.split(",")
        value = []
        for t1 in vv:
            phi = []
            for t2 in t1.strip().split():
                phi.append(t2.strip().upper())
            if len(phi) != self.length:
                raise ValueError("list requires " + str(self.length) + " elements")
            value.append(phi)

        if len(value):
            self.value = value

        return self.value


class StringValidator(Validator):
    def __init__(self, key, datatype, descr, default, valid_values=None):
        super().__init__(key, datatype, descr, default)
        self.valid_values = valid_values
        self.units = None

    def args(self):
        if self.valid_values:
            return self.valid_values
        else:
            return "free string"

    def validate(self, value_list, basedir=None):
        import re
        if type(value_list) is not list and type(value_list) is not tuple:
            if value_list is None:
                value_list = 'None'
            value_list = [value_list]
        for v in value_list:
            if self.valid_values and v not in self.valid_values:
                raise ValueError('Value must be one of: {}'.format(self.valid_values))

        value = ' '.join(value_list)
        value = value.strip()
        # Remove starting or ending quotes (single, double)
        value = re.sub('^"', '', value)
        value = re.sub('"$', '', value)
        value = re.sub("^'", '', value)
        value = re.sub("'$", '', value)
        return value


class TimestepValidator(Validator):
    def __init__(self, key, datatype, descr, default):
        super().__init__(key, datatype, descr, default)
        self.units = None

    def args(self):
        return "[ number (ps|ns|us) ]"

    def validate(self, value_list, basedir=None):
        if type(value_list) is not list and type(value_list) is not tuple:
            value_list = [value_list]

        try:
            if len(value_list) == 1:
                value = int(value_list[0])
            elif len(value_list) == 2:
                value = float(value_list[0])
                if value_list[1].lower() == "ps":
                    value = value * 1000 / 4  # TODO FIXME -- note the hard-coded assumption about the timestep
                if value_list[1].lower() == "ns":
                    value = value * 1000000 / 4
                if value_list[1].lower() == "us":
                    value = value * 1000000000 / 4
                value = int(value)
            else:
                raise ValueError()
        except:
            raise ValueError("Value must be >= 0 and may have a unit suffix [ps,ns,us]")
        return value


class ListValidator(Validator):
    def __init__(self, key, datatype, descr, default, valid_values):
        super().__init__(key, datatype, descr, default)
        self.valid_values = valid_values
        self.units = None

    def args(self):
        return str(self.valid_values)

    def validate(self, value_list, basedir=None):
        if type(value_list) is not list and type(value_list) is not tuple:
            value_list = [value_list]
        for value in value_list:
            if self.valid_values is not None and value not in self.valid_values:
                raise ValueError("Value must be one of " + str(self.valid_values))
        return value_list


class DictionaryValidator(Validator):
    def __init__(self, key, datatype, descr, default, key_type=None, val_type=None):
        super().__init__(key, datatype, descr, default)
        self.key_type = key_type
        self.val_type = val_type

    def args(self):
        return

    def validate(self, value, basedir=None):
        if type(value) is not dict:
            raise TypeError('Value is not a dictionary.')
        for k in value.keys():
            if self.key_type is not None and not isinstance(k, self.key_type):
                raise TypeError('Key {} is not of type {}'.format(k, self.key_type))
            if self.val_type is not None and not isinstance(value[k], self.val_type):
                raise TypeError('Value {} of key {} is not of type {}'.format(value[k], k, self.val_type))
        return value


class BinaryValidator(Validator):
    def __init__(self, key, type, descr, default):
        super().__init__(key, type, descr, default)
        self.units = None

    def args(self):
        return "[ on | off ]"

    def validate(self, value_list, basedir=None):
        if type(value_list) is not list and type(value_list) is not tuple:
            value_list = [value_list]
        value = value_list[0]

        try:
            if value:
                value = True
                return value
            if not value or value is None:
                value = False
                return value

            value = value.lower()

            if value == "yes" or value == "on" or value == "1" or value == "all" or value == "true":
                value = True
            elif value == "no" or value == "off" or value == "0" or value == "none" or value == "false":
                value = False
            else:
                raise ValueError("")
        except:
            raise ValueError("Value must be binary ( on|off )")
        return value


class FileValidator(Validator):
    def __init__(self, key, datatype, descr, default, exist=False, writable=False, multiple=False, check=None):
        super().__init__(key, datatype, descr, default)
        self.multiple = multiple
        self.must_exist = exist
        self.writable = writable
        self.check = check

    def args(self):
        dd = "file"
        if self.must_exist:
            dd = "input file"
        if self.writable:
            dd = "output file"
        if self.multiple:
            dd = "list of " + dd + "s"
        return "[ " + dd + " ]"

    def validate(self, value_list, basedir=None):
        if type(value_list) is not list and type(value_list) is not tuple:
            value_list = [value_list]

        l = []
        for value in value_list:
            if self.must_exist:
                found = False
                if basedir and os.path.isfile(os.path.join(basedir, value)):
                    value = os.path.join(basedir, value)
                    found = True
                if not found and os.path.isfile(value):
                    found = True
                if not found:
                    raise NameError("File '" + value + "' does not exist")
                if not os.access(value, os.R_OK):
                    raise NameError("File '" + value + "' cannot be read")
            if self.writable:
                if basedir:
                    value = os.path.join(basedir, value)
                try:
                    f = open(value, "a+")
                    f.close()
                except:
                    raise NameError("File '" + value + "' is not writable")
            l.append(value)

        # Test to see if the file exists
        if len(l) != 1 and not self.multiple:
            raise NameError("Value must be a single filename")
        if not self.multiple:
            value = l[0]
        else:
            value = l
        return value


class ValueValidator(Validator):
    def __init__(self, key, datatype, descr, default, realdatatype, valid_range, list_len=1, units=None):
        super().__init__(key, datatype, descr, default)
        self.valid_range = valid_range
        self.list_len = list_len
        self.units = units
        self.realdatatype = realdatatype

    def validate(self, value_list, basedir=None):
        valid_range = self.valid_range

        if type(value_list) is not list and type(value_list) is not tuple:
            value_list = [value_list]

        l = []
        if self.list_len != len(value_list):
            if self.list_len == 0:
                raise NameError("Value must be a scalar")
            else:
                raise NameError("Value must be a vector with " + str(self.list_len) + " elements")

        for value in value_list:
            if self.realdatatype == TYPE_FLOAT:
                strfudge = "a number"
            elif self.realdatatype == TYPE_INT:
                strfudge = "an integer"

            try:
                try:
                    if self.realdatatype == TYPE_FLOAT:
                        value = float(value)
                    if self.realdatatype == TYPE_INT:
                        value = int(value)
                except:
                    raise ValueError("Value is not " + strfudge)

                if valid_range == RANGE_POS and value <= 0:
                    raise ValueError("Value must be " + strfudge + " > 0")
                if valid_range == RANGE_0POS and value < 0:
                    raise ValueError("Value must be " + strfudge + " >=0")
                if valid_range == RANGE_NEG and value >= 0:
                    raise ValueError("Value must be " + strfudge + " < 0")
                if valid_range == RANGE_0NEG and value > 0:
                    raise ValueError("Value must be " + strfudge + " <=0")
            except NameError as e:
                raise NameError(e)
            if self.list_len == 1:
                l = value
            else:
                l.append(value)
        return l


'''class AttributeParser:
    def __init__(self, parent=None):
        if parent is not None:
            self.attributes = parent.attributes
            self.deprecated = parent.deprecated
            self.numAttributes = parent.numAttributes
            self.attributesOrder = parent.attributesOrder
        else:
            self.attributes = {}
            self.deprecated = {}
            self.numAttributes = 0
            self.attributesOrder = {}

    def addFile(self, key, datatype, descr, default, exist=False, writable=False, multiple=False, check=None):
        self.attributes[key] = FileValidator(datatype, descr, default, exist, writable, multiple, check)
        self.attributesOrder[key] = self.numAttributes
        self.numAttributes += 1

    def addString(self, key, datatype, descr, default, valid_values=None):
        self.attributes[key] = StringValidator(datatype, descr, default, valid_values)
        self.attributesOrder[key] = self.numAttributes
        self.numAttributes += 1

    def addTimestep(self, key, datatype, descr, default):
        self.attributes[key] = TimestepValidator(datatype, descr, default)
        self.attributesOrder[key] = self.numAttributes
        self.numAttributes += 1

    def addList(self, key, datatype, descr, default, valid_values):
        self.attributes[key] = ListValidator(datatype, descr, default, valid_values)
        self.attributesOrder[key] = self.numAttributes
        self.numAttributes += 1

    def addBinary(self, key, datatype, descr, default):
        self.attributes[key] = BinaryValidator(datatype, descr, default)
        self.attributesOrder[key] = self.numAttributes
        self.numAttributes += 1

    def addValue(self, key, datatype, descr, default, realdatatype, valid_range, list_len=1, units=None):
        self.attributes[key] = ValueValidator(datatype, descr, default, realdatatype, valid_range, list_len, units)
        self.attributesOrder[key] = self.numAttributes
        self.numAttributes += 1

    def addObject(self, key, datatype, descr, default, classname):
        self.attributes[key] = ObjectValidator(datatype, descr, default, classname)
        self.attributesOrder[key] = self.numAttributes
        self.numAttributes += 1

    def addBoolean(self, key, datatype, descr, default):
        self.attributes[key] = BooleanValidator(datatype, descr, default)
        self.attributesOrder[key] = self.numAttributes
        self.numAttributes += 1

    def addDeprecated(self, key, newkey=None):
        self.deprecated[key] = newkey

    def printDocString(self):
        docs = ''
        for k in sorted(self.attributesOrder.items(), key=lambda x: x[1]):
            docs += '{} : {}\n'.format(k[0], self.attributes[k[0]].docString())
        print(docs)

    def testDeprecation(self, key):
        if key in self.deprecated:
            newkey = self.deprecated[key]
            if not newkey:
                print(" Attribute '" + key + "'\t is deprecated and is no longer required")
            else:
                print(" Attribute '" + key + "'\t is deprecated and replaced by '" + newkey + "'")
                return newkey
        else:
            return key

    def getDefaultConfiguration(self):
        ret = {}
        for i in self.attributes:
            ret[i] = self.attributes[i].default
        return ret

    def validate(self, key, value, basedir=None):
        key = self.testDeprecation(key)
        if key not in self.attributes:
            strerror = "Attribute '" + key + "' not allowed in this class."
            match = get_close_matches(key, self.attributes)
            if match:
                strerror += " Try '" + match[0] + "'"
            raise ValueError(strerror)

        cmd = self.attributes[key]
        return cmd.validate(value, basedir=basedir)

    def setAttributes(self, obj):
        for a in self.attributes:
            obj.__dict__[a] = self.attributes[a].default'''



# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#


def compareVersions():
    from htmd.version import version
    from natsort import natsorted
    import os
    import time
    from os.path import expanduser

    __home = expanduser("~")
    __htmdconf = os.path.join(__home, ".htmd")
    if not os.path.exists(__htmdconf):
        try:
            os.makedirs(__htmdconf)
        except:
            print(
                "Unable to create {} folder. Will not check for new HTMD versions.".format(
                    __htmdconf
                )
            )
            return
    __file = os.path.join(__htmdconf, ".latestversion")

    # Check if one day has passed since last version check. If yes, get new version and write to file
    if (
        not os.path.isfile(__file)
        or time.time() > os.path.getmtime(__file) + 86400
        or os.stat(__file).st_size == 0
    ):
        _writeLatestVersionFile(__file)

    try:
        f = open(__file, "r")
    except:
        print(
            "Unable to open {} file for reading. Will not check for new HTMD versions.".format(
                __file
            )
        )
        return
    latestversions = f.readlines()
    f.close()

    if len(latestversions) != 2:
        print(
            "There is something wrong with your {} file. Will not check for new HTMD versions.".format(
                __file
            )
        )
        return

    currver = version()
    if currver != "unpackaged":
        if _is_stable(currver):
            pieces = latestversions[0].split()
            latest = pieces[0].strip()
            verstring = "stable"
        else:
            pieces = latestversions[1].split()
            latest = pieces[0].strip()
            verstring = "devel"
        pydeps = ""
        if len(pieces) > 1:
            pydeps = " python[{}]".format(pieces[1])
    elif currver == "unpackaged":
        pass
    else:
        return

    if currver != "unpackaged" and natsorted((latest, currver))[1] != currver:
        print(
            "New {verstring} HTMD version ({latest}{pydeps}) is available. You are currently on ({currver})."
            "There are several methods to update:"
            "    - Create a new conda env. using `conda create -n htmd{latest} htmd={latest} -c acellera -c psi4 -c conda-forge`"
            "    - Create a brand new conda installation and run `conda install htmd -c acellera -c psi4 -c conda-forge`"
            "    - Run: `conda update htmd -c acellera -c psi4 -c conda-forge` (NOT RECOMMENDED)"
            "".format(
                verstring=verstring, latest=latest, pydeps=pydeps, currver=currver
            )
        )
    else:
        if currver != "unpackaged":
            print("You are on the latest HTMD version ({}).".format(currver))
        else:
            from htmd.home import home

            print(
                "You are on the latest HTMD version ({} : {}).".format(currver, home())
            )

    print("")


def _writeLatestVersionFile(fname):
    import os

    try:
        f = open(fname, "w")
    except:
        print(
            "Unable to open {} file for writing. Will not check for new HTMD versions.".format(
                fname
            )
        )
        return

    try:
        from binstar_client.utils import get_server_api

        api = get_server_api(log_level=0)
        package = api.package("acellera", "htmd")
    except Exception as err:
        print(
            "Failed at checking latest conda version. ({})".format(type(err).__name__)
        )
        return

    stable, dev, stabledeps, devdeps = _release_version(package)

    f.write("{} {}\n{} {}".format(stable, ",".join(stabledeps), dev, ",".join(devdeps)))
    os.utime(fname, None)
    f.close()


def _release_version(package):

    versionlist = package["versions"]
    laststable = None
    lastdev = None
    for ver in versionlist[::-1]:  # Iterate in reverse due to sorting of conda versions
        if laststable is None and _is_stable(ver):
            laststable = ver
        elif lastdev is None:
            lastdev = ver
        if laststable and lastdev:
            break

    stabledeps = _release_python_dep(package, laststable)
    devdeps = _release_python_dep(package, lastdev)

    return laststable, lastdev, stabledeps, devdeps


def _is_stable(ver):
    import numpy as np

    if (
        np.mod(int(ver.split(".")[1]), 2) == 0
    ):  # Even versions are stable (i.e. 1.2.x 3.4.x etc)
        return True
    else:
        return False


def _release_python_dep(package, version, opersys=None):
    import platform

    if opersys is None:
        opersys = platform.system().lower()
        if opersys == "windows":
            opersys = "win"
    try:
        versions = []
        for f in package["files"]:
            if f["version"] == version and f["attrs"][
                "operatingsystem"
            ].lower().startswith(opersys.lower()):
                for d in f["dependencies"]["depends"]:
                    if d["name"].lower() == "python":
                        versions.append(d["specs"][0][1])
        if len(versions):
            return versions
        else:
            return " does not exist for your platform. Please create an issue on HTMD git issue tracker."
    except:
        return




# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#

from protocolinterface import ProtocolInterface, val
from htmd.apps.acemd import Acemd as Acemd2
from htmd.mdengine.acemd.acemd import Acemd, _Restraint, GroupRestraint, AtomRestraint
import os
import numpy as np
import logging
from htmd.config import _config
logger = logging.getLogger(__name__)


class Production(ProtocolInterface):
    """ Production protocol v6

        Production protocol for globular and membrane proteins. You can optionally define a flatbottom potential box and
        atom constraints for the production run.

        An Acemd class object is stored in the Production object which can be used to modify futher options.
        For documentation on further options see :class:`Acemd <htmd.mdengine.acemd.acemd.Acemd>`

        Parameters
        ----------
        runtime : float, default=0
            Running time of the simulation.
        timeunits : str, default='steps'
            Units for runtime. Can be 'steps', 'ns' etc.
        temperature : float, default=300
            Temperature of the thermostat in Kelvin
        fb_k : float, default=0
            Force constant of the flatbottom potential in kcal/mol/A^2. E.g. 5
        fb_reference : str, default='none'
            Reference selection to use as dynamic center of the flatbottom box.
        fb_selection : str, default='none'
            Selection of atoms to apply the flatbottom potential
        fb_box : list, default=[0, 0, 0, 0, 0, 0]
            Position of the flatbottom box in term of the reference center given as [xmin, xmax, ymin, ymax, zmin, zmax]
        useconstantratio : bool, default=False
            For membrane protein simulations set it to true so that the barostat does not modify the xy aspect ratio.
        useconstraints : bool, default=False
            Apply constraints to the production simulation, defined by the constraints parameter
        constraints : dict, default={}
            A dictionary of atomselections and values of the constraint to be applied (in kcal/mol/A^2). Atomselects must be mutually exclusive.
        adaptive : bool, default=False
            Set to True if making production runs for adaptive sampling.
    """
    def __init__(self, _version=_config['acemdversion']):
        super().__init__()
        self._version = _version
        self._arg('acemd', ':class:`Acemd2 <htmd.apps.acemd.Acemd>` or :class:`Acemd <htmd.mdengine.acemd.acemd.Acemd>`'
                           ' object', 'Acemd class object', None, val.Object([Acemd2, Acemd]))
        self._arg('runtime', 'float', 'Running time of the simulation.', 25000, val.Number(float, '0POS'))
        self._arg('timeunits', 'str', 'Units for runtime. Can be \'steps\', \'ns\' etc.', 'steps', val.String())
        self._arg('temperature', 'float', 'Temperature of the thermostat in Kelvin', 300, val.Number(float, 'ANY'))
        self._arg('fb_k', 'float', 'Force constant of the flatbottom potential in kcal/mol/A^2. E.g. 5', 0,
                  val.Number(float, 'ANY'))
        self._arg('fb_reference', 'str', 'Reference selection to use as dynamic center of the flatbottom box.',
                  'none', val.String())
        self._arg('fb_selection', 'str', 'Selection of atoms to apply the flatbottom potential', 'none', val.String())
        self._arg('fb_box', 'list', 'Position of the flatbottom box in term of the reference center given as '
                                    '[xmin, xmax, ymin, ymax, zmin, zmax]',
                  [0, 0, 0, 0, 0, 0], val.Number(float, 'ANY'), nargs=6)
        self._arg('useconstantratio', 'bool', 'For membrane protein simulations set it to true so that the barostat '
                                              'does not modify the xy aspect ratio.', False, val.Boolean())
        self._arg('useconstraints', 'bool', 'Apply constraints to the production simulation, defined by the '
                                            'constraints parameter', False, val.Boolean())
        self._arg('constraints', 'dict', 'A dictionary of atomselections and values of the constraint to be '
                                         'applied (in kcal/mol/A^2). Atomselects must be mutually exclusive.', {},
                  val.Dictionary(key_type=str))
        self._arg('adaptive', 'bool', 'Set to True if making production runs for adaptive sampling.', False,
                  val.Boolean())
        self._arg('restraints', 'list', 'A list of restraint objects. Only works with {}(_version=3),'
                                'see :class:`AtomRestraint <htmd.mdengine.acemd.acemd.AtomRestraint>` and'
                                ':class:`GroupRestraint <htmd.mdengine.acemd.acemd.GroupRestraint>`'
                                ')'.format(self.__class__.__name__), None, val.Object(_Restraint), nargs='*')

        if self._version == 2:
            self.acemd = Acemd2()
            self.acemd.binindex = None
            self.acemd.binvelocities = None
            self.acemd.bincoordinates = 'output.coor'
            self.acemd.extendedsystem = 'output.xsc'
            self.acemd.coordinates = None
            self.acemd.structure = None
            self.acemd.parameters = None
            self.acemd.temperature = None
            self.acemd.restart = 'on'
            self.acemd.restartfreq = '5000'
            self.acemd.outputname = 'output'
            self.acemd.xtcfile = 'output.xtc'
            self.acemd.xtcfreq = '25000'
            self.acemd.timestep = '4'
            self.acemd.rigidbonds = 'all'
            self.acemd.hydrogenscale = '4'
            self.acemd.switching = 'on'
            self.acemd.switchdist = '7.5'
            self.acemd.cutoff = '9'
            self.acemd.exclude = 'scaled1-4'
            self.acemd.scaling14 = '1.0'
            self.acemd.langevin = 'on'
            self.acemd.langevintemp = None
            self.acemd.langevindamping = '0.1'
            self.acemd.pme = 'on'
            self.acemd.pmegridspacing = '1.0'
            self.acemd.fullelectfrequency = '2'
            self.acemd.energyfreq = '5000'
            self.acemd.consref = None
            self.acemd.run = '$numsteps'
            self.acemd.TCL = ('''
set numsteps {NUMSTEPS}
set fb_refindex {{ {REFINDEX} }}
set fb_selindex {{ {SELINDEX} }}
set fb_box {{ {BOX} }}
set fb_K {KCONST}
#
''',
'''
proc flatbot1d {x xm xM fb_K} {
  set f 0
  if {$x < $xm} {
    set f [expr $fb_K*[expr $xm-$x]]
  }
  if {$x > $xM} {
    set f [expr $fb_K*[expr $xM-$x]]
  }
  return $f
}
proc calcforces_init {} {
  global ref sel fb_refindex fb_selindex
  berendsenpressure  off
  set ref [addgroup  $fb_refindex]
  set sel [addgroup  $fb_selindex]
}
proc calcforces {} {
  global ref sel fb_K fb_box
  loadcoords coords
##FLATBOTTOM
  if {$fb_K>0} {
    set r0 $coords($ref)
    set r1 $coords($sel)
    set dr  [vecsub $r1 $r0]
    set fx [flatbot1d [lindex $dr 0] [lindex $fb_box 0] [lindex $fb_box 1] $fb_K]
    set fy [flatbot1d [lindex $dr 1] [lindex $fb_box 2] [lindex $fb_box 3] $fb_K]
    set fz [flatbot1d [lindex $dr 2] [lindex $fb_box 4] [lindex $fb_box 5] $fb_K]
    #print "dr: $dr  fx: $fx fy: $fy fz: $fz"
    addforce $sel [list $fx $fy $fz]
  }
}
proc calcforces_endstep { } { }
''')
        elif self._version == 3:
            self.acemd = Acemd()
            self.acemd.binvelocities = None
            self.acemd.bincoordinates = 'output.coor'
            self.acemd.extendedsystem = 'output.xsc'
            self.acemd.coordinates = None
            self.acemd.structure = None
            self.acemd.parameters = None
            self.acemd.restart = 'on'
            self.acemd.trajectoryfile = 'output.xtc'
            self.acemd.trajectoryperiod = 25000
            self.acemd.timestep = 4
            self.acemd.switching = 'on'
            self.acemd.switchdistance = 7.5
            self.acemd.cutoff = 9
            self.acemd.thermostat = 'on'
            self.acemd.thermostatdamping = 0.1
            self.acemd.pme = 'on'
        else:
            raise ValueError('_version can not be {}. Choose either 2 or 3.'.format(self._version))

    def _findFiles(self, inputdir):
        # Tries to find default files if the given don't exist
        defaults = {'coordinates': ('structure.pdb',),
                    'structure': ('structure.psf', 'structure.prmtop'),
                    'parameters': ('parameters', 'structure.prmtop')}

        for field in defaults:
            userval = self.acemd.__dict__[field]
            if userval is not None and not os.path.exists(os.path.join(inputdir, userval)):
                raise RuntimeError('Could not locate file {} set by the user for argument '
                                   'Production.acemd.{}'.format(os.path.join(inputdir, userval), field))

            if self.acemd.__dict__[field] is None:
                for val in defaults[field]:
                    if os.path.exists(os.path.join(inputdir, val)):
                        self.acemd.__dict__[field] = val
                        break

            if userval is not None and self.acemd.__dict__[field] is not None and self.acemd.__dict__[field] != userval:
                logger.warning('Could not locate structure file {}. Using {} instead.'.format(
                    os.path.join(inputdir, userval), os.path.join(inputdir, self.acemd.__dict__[field])
                ))
            elif self.acemd.__dict__[field] is None:
                raise RuntimeError('Could not locate any {f:} file in {i:}. '
                                   'Please set the Production.acemd.{f:} property to '
                                   'point to the {f:} file'.format(f=field, i=inputdir))

            if self._version == 2:
                if self.useconstraints and self.acemd.consref is None:
                    self.acemd.consref = self.acemd.coordinates

    def _amberFixes(self):
        # AMBER specific fixes
        if self.acemd.structure.endswith('.prmtop'):
            self.acemd.parmfile = self.acemd.parameters
            self.acemd.parameters = None
            if self._version == 2:
                self.acemd.scaling14 = '0.8333333'
                self.acemd.amber = 'on'

    def _fb_potential2restraints(self, inputdir):
        from moleculekit.molecule import Molecule
        restraints = list()

        fb_box = np.array(self.fb_box)
        # convert fb_box to width
        width = list(np.concatenate(np.diff(np.array([fb_box[::2], fb_box[1::2]]), axis=0)))

        # If fb_box is not symmetrical
        if not np.all(fb_box[::2] == -fb_box[1::2]):
            # convert fb_box and fb_reference to fbcentre and width
            mol = Molecule(os.path.join(inputdir, self.acemd.structure))
            mol.read(os.path.join(inputdir, self.acemd.coordinates))
            fb_refcentre = mol.get('coords', sel=self.fb_reference).mean(axis=0).squeeze()

            fbcentre = list(np.around(np.mean(np.array([fb_box[::2], fb_box[1::2]]), axis=0) + fb_refcentre, 3))
            restraints.append(GroupRestraint(self.fb_selection, width, [(self.fb_k, 0)], fbcentre=fbcentre))
        else:
            restraints.append(GroupRestraint(self.fb_selection, width, [(self.fb_k, 0)], fbcentresel=self.fb_reference))

        return restraints

    def _constraints2restraints(self):

        restraints = list()
        for constr in sorted(self.constraints):
            restraints.append(AtomRestraint(constr, 0, [(self.constraints[constr], 0)]))

        return restraints

    def write(self, inputdir, outputdir):
        """ Writes the production protocol and files into a folder.

        Parameters
        ----------
        inputdir : str
            Path to a directory containing the files produced by a equilibration process.
        outputdir : str
            Directory where to write the production setup files.
        """
        from moleculekit.molecule import Molecule

        # Do version consistency check
        if (self._version == 2 and not isinstance(self.acemd, Acemd2)) and \
                (self._version == 3 and not isinstance(self.acemd, Acemd)):
            raise RuntimeError('Acemd object version ({}) inconsistent with protocol version at instantiation '
                               '({})'.format(type(self.acemd), self._version))

        self._findFiles(inputdir)
        self._amberFixes()

        if self._version == 2:
            self.acemd.temperature = str(self.temperature)
            self.acemd.langevintemp = str(self.temperature)
        elif self._version == 3:
            self.acemd.temperature = self.temperature
            self.acemd.thermostattemperature = self.temperature

        from htmd.units import convert
        numsteps = convert(self.timeunits, 'timesteps', self.runtime, timestep=self.acemd.timestep)
        if self._version == 3:
            self.acemd.run = str(numsteps)

        pdbfile = os.path.join(inputdir, self.acemd.coordinates)
        inmol = Molecule(pdbfile)

        from htmd.builder.builder import detectCisPeptideBonds
        detectCisPeptideBonds(inmol)

        if np.any(inmol.atomselect('lipids')) and not self.useconstantratio:
            logger.warning('Lipids detected in input structure. We highly recommend setting useconstantratio=True '
                           'for membrane simulations.')

        if self._version == 2:
            if self.restraints:
                raise RuntimeWarning('restraints are only available on {}(_version=3)'.format(self.__class__.__name__))
            if self.fb_k > 0:  # use TCL only for flatbottom
                self.acemd.tclforces = 'on'
                if isinstance(self.acemd.TCL, tuple):
                    tcl = list(self.acemd.TCL)
                    tcl[0] = tcl[0].format(NUMSTEPS=numsteps, TEMPERATURE=self.temperature, KCONST=self.fb_k,
                                           REFINDEX=' '.join(map(str, inmol.get('index', self.fb_reference))),
                                           SELINDEX=' '.join(map(str, inmol.get('index', self.fb_selection))),
                                           BOX=' '.join(map(str, self.fb_box)))
                    self.acemd.TCL = tcl[0] + tcl[1]
                else:
                    logger.warning('{} default TCL was already formatted.'.format(self.__class__.__name__))
            else:
                self.acemd.TCL = 'set numsteps {NUMSTEPS}\n'.format(NUMSTEPS=numsteps)
            if self.useconstraints:
                # Turn on constraints
                self.acemd.constraints = 'on'
                self.acemd.constraintscaling = '1.0'
            else:
                if len(self.constraints) != 0:
                    logger.warning('You have setup constraints to {} but constraints are turned off. '
                                   'If you want to use constraints, define '
                                   'useconstraints=True'.format(self.constraints))
        elif self._version == 3:
            if self.restraints is not None:
                logger.info('Using user-provided restraints and ignoring constraints and fb_potential')
                self.acemd.restraints = self.restraints
            else:
                restraints = list()
                if self.fb_k > 0:
                    logger.warning('Converting fb_potential to restraints. This is a convenience '
                                   'functional conversion. We recommend start using restraints with '
                                   '{}(_version=3)'.format(self.__class__.__name__))
                    restraints += self._fb_potential2restraints(inputdir)
                if self.useconstraints:
                    logger.warning('Converting constraints to restraints. This is a convenience '
                                   'functional conversion. We recommend start using restraints with '
                                   '{}(_version=3)'.format(self.__class__.__name__))
                    restraints += self._constraints2restraints()
                else:
                    if len(self.constraints) != 0:
                        logger.warning('You have setup constraints to {} but constraints are turned off. '
                                       'If you want to use constraints, define '
                                       'useconstraints=True'.format(self.constraints))
                if len(restraints) != 0:
                    self.acemd.restraints = restraints

        if self.useconstantratio:
            self.acemd.useconstantratio = 'on'

        if self.adaptive:
            self.acemd.binvelocities = None

        self.acemd.setup(inputdir, outputdir, overwrite=True)

        if self._version == 2:
            # Adding constraints by writing them to the consref file
            if self.useconstraints:
                inconsreffile = os.path.join(inputdir, self.acemd.consref)
                consrefmol = Molecule(inconsreffile)
                consrefmol.set('occupancy', 0)
                consrefmol.set('beta', 0)
                if len(self.constraints) == 0:
                    raise RuntimeError('You have set the production to use constraints (useconstraints=True), but have '
                                       'not defined any constraints (constraints={}).')
                else:
                    for sel in self.constraints:
                        consrefmol.set('beta', self.constraints[sel], sel)
                outconsreffile = os.path.join(outputdir, self.acemd.consref)
                consrefmol.write(outconsreffile)

    def addConstraint(self, atomselect, factor=1):
        """ Convenience function for adding a new constraint to existing constraints.

        Parameters
        ----------
        atomselect : str
            Atom selection of atoms we want to constrain
        factor : float
            The scaling factor of the constraints applied to the atoms

        Example
        -------
        >>> eq.addConstraint('chain X', 0.3)
        """
        self.constraints[atomselect] = factor


import unittest
class _TestProduction(unittest.TestCase):
    def _cutfirstline(self, infile, outfile):
        # Cut out the first line of prmtop which has a build date in it
        with open(infile, 'r') as fin:
            data = fin.read().splitlines(True)
        with open(outfile, 'w') as fout:
            fout.writelines(data[1:])

    def _compareResultFolders(self, compare, tmpdir, pid):
        from glob import glob
        import os
        import filecmp

        ignore_ftypes = ('.log', '.txt')
        files = []
        deletefiles = []
        for f in glob(os.path.join(compare, '*')):
            fname = os.path.basename(f)
            if os.path.splitext(f)[1] in ignore_ftypes:
                continue
            if f.endswith('prmtop'):
                self._cutfirstline(f, os.path.join(compare, fname + '.mod'))
                self._cutfirstline(os.path.join(tmpdir, fname), os.path.join(tmpdir, fname + '.mod'))
                files.append(os.path.basename(f) + '.mod')
                deletefiles.append(os.path.join(compare, fname + '.mod'))
            else:
                files.append(os.path.basename(f))

        match, mismatch, error = filecmp.cmpfiles(tmpdir, compare, files, shallow=False)
        if len(mismatch) != 0 or len(error) != 0 or len(match) != len(files):
            raise RuntimeError(
                'Different results produced by amber.build for test {} between {} and {} in files {}.'.format(pid, compare, tmpdir, mismatch))

        for f in deletefiles:
            os.remove(f)

    def test_acemd2(self):
        from htmd.util import tempname
        from htmd.home import home
        from glob import glob
        import os

        pd = Production(_version=2)
        pd.runtime = 4
        pd.timeunits = 'ns'
        pd.temperature = 300
        pd.useconstraints = True
        pd.constraints = {'protein and name CA': 1, 'protein and noh and not name CA': 0.1}
        pd.fb_reference = 'protein and name CA'
        pd.fb_selection = 'resname MOL and noh'
        pd.fb_box = [-21, 21, -19, 19, 29, 30]
        pd.fb_k = 5
        tmpdir = tempname()
        pd.write(home(dataDir=os.path.join('test-protocols', 'equilibration', 'acemd2', 'protLig', 'postrun')), tmpdir)

        # Compare with reference
        refdir = home(dataDir=os.path.join('test-protocols', 'production', 'acemd2', 'protLig', 'prerun'))
        files = [os.path.basename(f) for f in glob(os.path.join(refdir, '*'))]
        self._compareResultFolders(refdir, tmpdir, 'protLig')

    def test_acemd3(self):
        from htmd.util import tempname
        from htmd.home import home
        from glob import glob
        import os

        pd = Production(_version=3)
        pd.runtime = 4
        pd.timeunits = 'ns'
        pd.temperature = 300
        pd.useconstraints = True
        pd.constraints = {'protein and name CA': 1, 'protein and noh and not name CA': 0.1}
        pd.fb_reference = 'protein and name CA'
        pd.fb_selection = 'resname MOL and noh'
        pd.fb_box = [-21, 21, -19, 19, 29, 30]
        pd.fb_k = 5
        tmpdir = tempname()
        pd.write(home(dataDir=os.path.join('test-protocols', 'equilibration', 'acemd3', 'protLig', 'postrun')), tmpdir)

        # Compare with reference
        refdir = home(dataDir=os.path.join('test-protocols', 'production', 'acemd3', 'protLig', 'prerun'))
        files = [os.path.basename(f) for f in glob(os.path.join(refdir, '*'))]
        self._compareResultFolders(refdir, tmpdir, 'protLig')

    @unittest.skipUnless('ACE3ARG' in os.environ, 'Untrusted PR')
    def test_run_water(self):
        from htmd.util import tempname
        from htmd.home import home
        from glob import glob
        from subprocess import check_output
        import subprocess
        import shutil
        import os

        acemd3exe = shutil.which('acemd3', mode=os.X_OK)
        if not acemd3exe:
            raise NameError('Could not find acemd3, or no execute permissions are given')

        for system in ['amber-equil-completed', 'charmm-equil-completed']:
            pd = Production(_version=3)
            pd.runtime = 5
            pd.timeunits = 'steps'
            pd.temperature = 300
            pd.constraints = {}
            # Set these down for tiny box size of water
            pd.acemd.cutoff = 3
            pd.acemd.switchdistance = 2
            ######
            tmpdir = tempname()
            pd.write(home(dataDir=os.path.join('test-acemd', 'tiny-water', system)), tmpdir)
            try:
                res = check_output(['acemd3', '--platform', 'CPU', os.getenv('ACE3ARG')], cwd=tmpdir)
            except subprocess.CalledProcessError as exc:
                assert False, f'Failed to run due to error: {exc}\n\n ---> Error log:\n\n{exc.output.decode("ascii")}'
            res = res.decode('utf-8').strip()
            print(res)
            assert res.endswith('Completed simulation!'), 'Failed at system ' + system


if __name__ == "__main__":
    unittest.main(verbosity=2)





# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
from htmd.apps.acemd import Acemd as Acemd2
from htmd.mdengine.acemd.acemd import Acemd, _Restraint, GroupRestraint, AtomRestraint
from htmd.config import _config
from protocolinterface import ProtocolInterface, val
import os
import numpy as np
import logging
logger = logging.getLogger(__name__)


class Equilibration(ProtocolInterface):
    """ Equilibration protocol v2

        Equilibration protocol for globular and membrane proteins
        It includes a flatbottom potential box to retrain a ligand
        for example within this box.

        Parameters
        ----------
        runtime : float, default=0
            Running time of the simulation.
        timeunits : str, default='steps'
            Units for time arguments. Can be 'steps', 'ns' etc.
        temperature : float, default=300
            Temperature of the thermostat in Kelvin
        fb_k : float, default=0
            Force constant of the flatbottom potential in kcal/mol/A^2. E.g. 5
        fb_reference : str, default='none'
            Reference selection to use as dynamic center of the flatbottom box.
        fb_selection : str, default='none'
            Selection of atoms to apply the flatbottom potential
        fb_box : list, default=[0, 0, 0, 0, 0, 0]
            Position of the flatbottom box in term of the reference center given as [xmin, xmax, ymin, ymax, zmin, zmax]
        useconstantratio : bool, default=False
            For membrane protein simulations set it to true so that the barostat does not modify the xy aspect ratio.
        constraints : dict, default={'protein and name CA': 1, 'protein and noh and not name CA': 0.1}
            A dictionary of atomselections and values of the constraint to be applied (in kcal/mol/A^2). Atomselects must be mutually exclusive.
        nvtsteps : int, default=500
            Number of initial steps to apply NVT in units of 4fs.
        constraintsteps : int, default=None
            Number of initial steps to apply constraints in units of 4fs. Defaults to half the simulation time.

        Example
        -------
        >>> from htmd.protocols.equilibration_v2 import Equilibration
        >>> md = Equilibration()
        >>> md.runtime = 4
        >>> md.timeunits = 'ns'
        >>> md.temperature = 300
        >>> md.useconstantratio = True  # only for membrane sims
        >>> # this is only needed for setting the flatbottom potential, otherwise remove it
        >>> md.fb_reference = 'protein and resid 293'
        >>> md.fb_selection = 'segname L and noh'
        >>> md.fb_box = [-25, 25, -25, 25, 43, 45]
        >>> md.fb_k = 5
        >>> md.write('./build','./equil')
    """
    def __init__(self, _version=_config['acemdversion']):
        super().__init__()
        self._version = _version
        self._arg('acemd', ':class:`Acemd2 <htmd.apps.acemd.Acemd>` or :class:`Acemd <htmd.mdengine.acemd.acemd.Acemd>`'
                           ' object', 'Acemd class object', None, val.Object([Acemd2, Acemd]))
        self._arg('runtime', 'float', 'Running time of the simulation.', 25000, val.Number(float, '0POS'))
        self._arg('timeunits', 'str', 'Units for time arguments. Can be \'steps\', \'ns\' etc.', 'steps', val.String())
        self._arg('temperature', 'float', 'Temperature of the thermostat in Kelvin', 300, val.Number(float, 'ANY'))

        self._arg('fb_k', 'float', 'Force constant of the flatbottom potential in kcal/mol/A^2. E.g. 5', 0,
                  val.Number(float, 'ANY'))
        self._arg('fb_reference', 'str', 'Reference selection to use as dynamic center of the flatbottom box.', 'none',
                  val.String())
        self._arg('fb_selection', 'str', 'Selection of atoms to apply the flatbottom potential', 'none', val.String())
        self._arg('fb_box', 'list', 'Position of the flatbottom box in term of the reference center given as '
                                    '[xmin, xmax, ymin, ymax, zmin, zmax]',
                  [0, 0, 0, 0, 0, 0], val.Number(float, 'ANY'), nargs=6)

        self._arg('constraints', 'dict', 'A dictionary of atomselections and values of the constraint to be '
                                         'applied (in kcal/mol/A^2). Atomselects must be mutually exclusive.',
                  {'protein and noh and not name CA': 0.1, 'protein and name CA': 1}, val.Dictionary(key_type=str))
        self._arg('useconstantratio', 'bool', 'For membrane protein simulations set it to true so that the barostat '
                                              'does not modify the xy aspect ratio.', False, val.Boolean())

        self._arg('nvtsteps', 'int', 'Number of initial steps to apply NVT in units of 4fs.', 500,
                  val.Number(int, 'ANY'))
        self._arg('constraintsteps', 'int', 'Number of initial steps to apply constraints in units of 4fs. Defaults '
                                            'to half the simulation time.', None, val.Number(int, 'ANY'))
        self._arg('restraints', 'list', 'A list of restraint objects. Only works with {}(_version=3),'
                                        'see :class:`AtomRestraint <htmd.mdengine.acemd.acemd.AtomRestraint>` and'
                                        ':class:`GroupRestraint <htmd.mdengine.acemd.acemd.GroupRestraint>`'
                                        ')'.format(self.__class__.__name__), None, val.Object(_Restraint), nargs='*')

        if self._version == 2:
            self.acemd = Acemd2()
            self.acemd.coordinates = None
            self.acemd.structure = None
            self.acemd.parameters = None
            self.acemd.temperature = '$temperature'
            self.acemd.restart = 'on'
            self.acemd.restartfreq = '5000'
            self.acemd.outputname = 'output'
            self.acemd.xtcfile = 'output.xtc'
            self.acemd.xtcfreq = '25000'
            self.acemd.timestep = '4'
            self.acemd.rigidbonds = 'all'
            self.acemd.hydrogenscale = '4'
            self.acemd.switching = 'on'
            self.acemd.switchdist = '7.5'
            self.acemd.cutoff = '9'
            self.acemd.exclude = 'scaled1-4'
            self.acemd.scaling14 = '1.0'
            self.acemd.langevin = 'on'
            self.acemd.langevintemp = '$temperature'
            self.acemd.langevindamping = '1'
            self.acemd.pme = 'on'
            self.acemd.pmegridspacing = '1.0'
            self.acemd.fullelectfrequency = '2'
            self.acemd.energyfreq = '1000'
            self.acemd.constraints = 'on'
            self.acemd.consref = None
            self.acemd.constraintscaling = '1.0'
            self.acemd.berendsenpressure = 'on'
            self.acemd.berendsenpressuretarget = '1.01325'
            self.acemd.berendsenpressurerelaxationtime = '800'
            self.acemd.tclforces = 'on'
            self.acemd.minimize = '500'
            self.acemd.run = '$numsteps'
            self.acemd.TCL = ('''
set numsteps {NUMSTEPS}
set temperature {TEMPERATURE}
set nvtsteps {NVTSTEPS}
set constraintsteps {CONSTRAINTSTEPS}
set fb_refindex {{ {REFINDEX} }}
set fb_selindex {{ {SELINDEX} }}
set fb_box {{ {BOX} }}
set fb_K {KCONST}
#
''',
'''
proc flatbot1d {x xm xM fb_K} {
  set f 0
  if {$x < $xm} {
    set f [expr $fb_K*[expr $xm-$x]]
  }
  if {$x > $xM} {
    set f [expr $fb_K*[expr $xM-$x]]
  }
  return $f
}
proc calcforces_init {} {
  global ref sel fb_refindex fb_selindex
  berendsenpressure  off
  set ref [addgroup  $fb_refindex]
  set sel [addgroup  $fb_selindex]
}
proc calcforces {} {
  global ref sel numsteps fb_K fb_box nvtsteps constraintsteps
  loadcoords coords
##FLATBOTTOM
  if {$fb_K>0} {
    set r0 $coords($ref)
    set r1 $coords($sel)
    set dr  [vecsub $r1 $r0]
    set fx [flatbot1d [lindex $dr 0] [lindex $fb_box 0] [lindex $fb_box 1] $fb_K]
    set fy [flatbot1d [lindex $dr 1] [lindex $fb_box 2] [lindex $fb_box 3] $fb_K]
    set fz [flatbot1d [lindex $dr 2] [lindex $fb_box 4] [lindex $fb_box 5] $fb_K]
    #print "dr: $dr  fx: $fx fy: $fy fz: $fz"
    addforce $sel [list $fx $fy $fz]
  }
##EQUIL
  set step [ getstep ]
  if { $step > $nvtsteps } {
    berendsenpressure  on
  } else {
    berendsenpressure  off
  }
  if { $step > $constraintsteps } {
    constraintscaling 0
  } else {
    constraintscaling [expr 1 - 0.95*$step/$constraintsteps]
  }
}
proc calcforces_endstep { } { }
''')
        elif self._version == 3:
            self.acemd = Acemd()
            self.acemd.coordinates = None
            self.acemd.structure = None
            self.acemd.parameters = None
            self.acemd.restart = 'on'
            self.acemd.trajectoryfile = 'output.xtc'
            self.acemd.trajectoryperiod = 25000
            self.acemd.timestep = 4
            self.acemd.switching = 'on'
            self.acemd.switchdistance = 7.5
            self.acemd.cutoff = 9
            self.acemd.thermostat = 'on'
            self.acemd.thermostatdamping = 1
            self.acemd.pme = 'on'
            self.acemd.barostat = 'on'
            self.acemd.barostatpressure = 1.01325
            self.acemd.minimize = 500
        else:
            raise ValueError('_version can not be {}. Choose either 2 or 3.'.format(self._version))

    def _findFiles(self, inputdir):
        # Tries to find default files if the given don't exist
        defaults = {'coordinates': ('structure.pdb', ),
                    'structure': ('structure.psf', 'structure.prmtop'),
                    'parameters': ('parameters', 'structure.prmtop')}

        for field in defaults:
            userval = self.acemd.__dict__[field]
            if userval is not None and not os.path.exists(os.path.join(inputdir, userval)):
                self.acemd.__dict__[field] = None

            if self.acemd.__dict__[field] is None:
                for val in defaults[field]:
                    if os.path.exists(os.path.join(inputdir, val)):
                        self.acemd.__dict__[field] = val
                        break

            if userval is not None and self.acemd.__dict__[field] is not None and self.acemd.__dict__[field] != userval:
                logger.warning('Could not locate structure file {}. Using {} instead.'.format(
                    os.path.join(inputdir, userval), os.path.join(inputdir, self.acemd.__dict__[field])
                ))
            elif self.acemd.__dict__[field] is None:
                raise RuntimeError('Could not locate any {f:} file in {i:}. '
                                   'Please set the {name:}.acemd.{f:} property to '
                                   'point to the {f:} file'.format(f=field, i=inputdir, name=self.__class__.__name__))

        if self._version == 2:
            if self.acemd.consref is None:
                self.acemd.consref = self.acemd.coordinates

    def _amberFixes(self):
        # AMBER specific fixes
        if self.acemd.parameters.endswith('structure.prmtop'):
            self.acemd.parmfile = self.acemd.parameters
            self.acemd.parameters = None
            if self._version == 2:
                self.acemd.scaling14 = '0.8333333'
                self.acemd.amber = 'on'

    def _constraints2restraints(self, constrsteps):

        restraints = list()
        for constr in sorted(self.constraints):
            restraints.append(AtomRestraint(constr, 0, [(self.constraints[constr], 0), (0, constrsteps)]))

        return restraints

    def _fb_potential2restraints(self, inputdir):
        from moleculekit.molecule import Molecule
        restraints = list()

        fb_box = np.array(self.fb_box)
        # convert fb_box to width
        width = list(np.concatenate(np.diff(np.array([fb_box[::2], fb_box[1::2]]), axis=0)))

        # If fb_box is not symmetrical
        if not np.all(fb_box[::2] == -fb_box[1::2]):
            # convert fb_box and fb_reference to fbcentre and width
            mol = Molecule(os.path.join(inputdir, self.acemd.structure))
            mol.read(os.path.join(inputdir, self.acemd.coordinates))
            fb_refcentre = mol.get('coords', sel=self.fb_reference).mean(axis=0).squeeze()

            fbcentre = list(np.around(np.mean(np.array([fb_box[::2], fb_box[1::2]]), axis=0) + fb_refcentre, 3))
            restraints.append(GroupRestraint(self.fb_selection, width, [(self.fb_k, 0)], fbcentre=fbcentre))
        else:
            restraints.append(GroupRestraint(self.fb_selection, width, [(self.fb_k, 0)], fbcentresel=self.fb_reference))

        return restraints

    def write(self, inputdir, outputdir):
        """ Write the equilibration protocol

        Writes the equilibration protocol and files into a folder for execution
        using files inside the inputdir directory

        Parameters
        ----------
        inputdir : str
            Path to a directory containing the files produced by a build process.
        outputdir : str
            Directory where to write the equilibration setup files.

        Examples
        --------
        >>> md = Equilibration()
        >>> md.write('./build','./equil')
        """

        from moleculekit.molecule import Molecule

        # Do version consistency check
        if (self._version == 2 and not isinstance(self.acemd, Acemd2)) and \
                (self._version == 3 and not isinstance(self.acemd, Acemd)):
            raise RuntimeError('Acemd object version ({}) inconsistent with protocol version at instantiation '
                               '({})'.format(type(self.acemd), self._version))

        self._findFiles(inputdir)
        self._amberFixes()

        from htmd.units import convert
        numsteps = convert(self.timeunits, 'timesteps', self.runtime, timestep=self.acemd.timestep)
        if self._version == 3:
            self.acemd.temperature = self.temperature
            self.acemd.thermostattemperature = self.temperature
            self.acemd.run = str(numsteps)

        pdbfile = os.path.join(inputdir, self.acemd.coordinates)
        inmol = Molecule(pdbfile)

        from htmd.builder.builder import detectCisPeptideBonds
        detectCisPeptideBonds(inmol)

        if np.any(inmol.atomselect('lipids')) and not self.useconstantratio:
            logger.warning('Lipids detected in input structure. We highly recommend setting useconstantratio=True '
                           'for membrane simulations.')

        if self.constraintsteps is None:
            constrsteps = int(numsteps / 2)
        else:
            constrsteps = int(self.constraintsteps)

        if self._version == 2:
            if self.restraints:
                raise RuntimeWarning('restraints are only available on {}(_version=3)'.format(self.__class__.__name__))
            if isinstance(self.acemd.TCL, tuple):
                tcl = list(self.acemd.TCL)
                tcl[0] = tcl[0].format(NUMSTEPS=numsteps, KCONST=self.fb_k,
                                       REFINDEX=' '.join(map(str, inmol.get('index', self.fb_reference))),
                                       SELINDEX=' '.join(map(str, inmol.get('index', self.fb_selection))),
                                       BOX=' '.join(map(str, self.fb_box)),
                                       NVTSTEPS=self.nvtsteps, CONSTRAINTSTEPS=constrsteps, TEMPERATURE=self.temperature)
                self.acemd.TCL = tcl[0] + tcl[1]
            else:
                logger.warning('{} default TCL was already formatted.'.format(self.__class__.__name__))
        elif self._version == 3:
            if self.restraints is not None:
                logger.info('Using user-provided restraints and ignoring constraints and fb_potential')
                self.acemd.restraints = self.restraints
            else:
                logger.warning('Converting constraints and fb_potential to restraints. This is a convenience '
                               'functional conversion. We recommend start using restraints with '
                               '{}(_version=3)'.format(self.__class__.__name__))
                restraints = list()
                # convert constraints to restraints and add them
                if self.constraints is not None:
                    restraints += self._constraints2restraints(constrsteps)
                # convert fb_potential to restraints and add them
                if self.fb_k > 0:
                    restraints += self._fb_potential2restraints(inputdir)
                self.acemd.restraints = restraints

        if ((self._version == 2) and self.acemd.celldimension is None and self.acemd.extendedsystem is None) or \
            ((self._version == 3) and self.acemd.boxsize is None and self.acemd.extendedsystem is None):
            coords = inmol.get('coords', sel='water')
            if coords.size == 0:  # It's a vacuum simulation
                coords = inmol.get('coords', sel='all')
                dim = np.max(coords, axis=0) - np.min(coords, axis=0)
                dim += 12.
            else:
                dim = np.max(coords, axis=0) - np.min(coords, axis=0)
            if self._version == 2:
                self.acemd.celldimension = '{} {} {}'.format(dim[0], dim[1], dim[2])
            else:
                self.acemd.boxsize = '{} {} {}'.format(dim[0], dim[1], dim[2])

        if self.useconstantratio:
            self.acemd.useconstantratio = 'on'

        self.acemd.setup(inputdir, outputdir, overwrite=True)

        if self._version == 2:
            # Adding constraints by writing them to the consref file
            inconsreffile = os.path.join(inputdir, self.acemd.consref)
            consrefmol = Molecule(inconsreffile)
            consrefmol.set('occupancy', 0)
            consrefmol.set('beta', 0)
            if len(self.constraints) == 0:
                raise RuntimeError('You have not defined any constraints for the {} ('
                                   'constraints={{}}).'.format(self.__class__.__name__))
            else:
                for sel in self.constraints:
                    consrefmol.set('beta', self.constraints[sel], sel)
            outconsreffile = os.path.join(outputdir, self.acemd.consref)
            consrefmol.write(outconsreffile)

    def addConstraint(self, atomselect, factor=1):
        """ Convenience function for adding a new constraint to existing constraints.

        Parameters
        ----------
        atomselect : str
            Atom selection of atoms we want to constrain
        factor : float
            The scaling factor of the constraints applied to the atoms

        Example
        -------
        >>> eq.addConstraint('chain X', 0.3)
        """
        self.constraints[atomselect] = factor


import unittest
class _TestEquilibration(unittest.TestCase):
    def _cutfirstline(self, infile, outfile):
        # Cut out the first line of prmtop which has a build date in it
        with open(infile, 'r') as fin:
            data = fin.read().splitlines(True)
        with open(outfile, 'w') as fout:
            fout.writelines(data[1:])

    def _compareResultFolders(self, compare, tmpdir, pid):
        from glob import glob
        import os
        import filecmp

        ignore_ftypes = ('.log', '.txt')
        files = []
        deletefiles = []
        for f in glob(os.path.join(compare, '*')):
            fname = os.path.basename(f)
            if os.path.splitext(f)[1] in ignore_ftypes:
                continue
            if f.endswith('prmtop'):
                self._cutfirstline(f, os.path.join(compare, fname + '.mod'))
                self._cutfirstline(os.path.join(tmpdir, fname), os.path.join(tmpdir, fname + '.mod'))
                files.append(os.path.basename(f) + '.mod')
                deletefiles.append(os.path.join(compare, fname + '.mod'))
            else:
                files.append(os.path.basename(f))

        match, mismatch, error = filecmp.cmpfiles(tmpdir, compare, files, shallow=False)
        if len(mismatch) != 0 or len(error) != 0 or len(match) != len(files):
            raise RuntimeError(
                'Different results produced by amber.build for test {} between {} and {} in files {}.'.format(pid, compare, tmpdir, mismatch))

        for f in deletefiles:
            os.remove(f)

    def test_acemd2(self):
        from htmd.util import tempname
        from htmd.home import home
        from glob import glob
        import os

        eq = Equilibration(_version=2)
        eq.runtime = 4
        eq.timeunits = 'ns'
        eq.temperature = 300
        # keep protein on the same place during all equilibration
        from htmd.units import convert
        eq.constraintsteps = convert(eq.timeunits, 'timesteps', eq.runtime, timestep=eq.acemd.timestep)
        eq.constraints = {'protein and name CA': 1, 'protein and noh and not name CA': 0.1}
        eq.fb_reference = 'protein and name CA'
        eq.fb_selection = 'resname MOL and noh'
        eq.fb_box = [-21, 21, -19, 19, 29, 30]
        eq.fb_k = 5
        tmpdir = tempname()
        eq.write(home(dataDir=os.path.join('test-protocols', 'build', 'protLig')), tmpdir)

        # Compare with reference
        refdir = home(dataDir=os.path.join('test-protocols', 'equilibration', 'acemd2', 'protLig', 'prerun'))
        files = [os.path.basename(f) for f in glob(os.path.join(refdir, '*'))]
        self._compareResultFolders(refdir, tmpdir, 'protLig')

    def test_acemd3(self):
        from htmd.util import tempname
        from htmd.home import home
        from glob import glob
        import os

        eq = Equilibration(_version=3)
        eq.runtime = 4
        eq.timeunits = 'ns'
        eq.temperature = 300
        # keep protein on the same place during all equilibration
        from htmd.units import convert
        eq.constraintsteps = convert(eq.timeunits, 'timesteps', eq.runtime, timestep=eq.acemd.timestep)
        eq.constraints = {'protein and name CA': 1, 'protein and noh and not name CA': 0.1}
        eq.fb_reference = 'protein and name CA'
        eq.fb_selection = 'resname MOL and noh'
        eq.fb_box = [-21, 21, -19, 19, 29, 30]
        eq.fb_k = 5
        tmpdir = tempname()
        eq.write(home(dataDir=os.path.join('test-protocols', 'build', 'protLig')), tmpdir)

        # Compare with reference
        refdir = home(dataDir=os.path.join('test-protocols', 'equilibration', 'acemd3', 'protLig', 'prerun'))
        files = [os.path.basename(f) for f in glob(os.path.join(refdir, '*'))]
        self._compareResultFolders(refdir, tmpdir, 'protLig')

    @unittest.skipUnless('ACE3ARG' in os.environ, 'Untrusted PR')
    def test_run_water(self):
        from htmd.util import tempname
        from htmd.home import home
        from glob import glob
        import subprocess
        from subprocess import check_output
        import shutil
        import os

        acemd3exe = shutil.which('acemd3', mode=os.X_OK)
        if not acemd3exe:
            raise NameError('Could not find acemd3, or no execute permissions are given')

        for system in ['amber-build', 'charmm-build']:
            eq = Equilibration(_version=3)
            eq.runtime = 5
            eq.timeunits = 'steps'
            eq.temperature = 300
            eq.constraints = {}
            # Set these down for tiny box size of water
            eq.acemd.cutoff = 3
            eq.acemd.switchdistance = 2
            eq.acemd.minimize = 50  # Do few steps to finish fast
            ######
            tmpdir = tempname()
            eq.write(home(dataDir=os.path.join('test-acemd', 'tiny-water', system)), tmpdir)
            try:
                res = check_output(['acemd3', '--platform', 'CPU', os.getenv('ACE3ARG')], cwd=tmpdir)
            except subprocess.CalledProcessError as exc:
                assert False, f'Failed to run due to error: {exc}\n\n ---> Error log:\n\n{exc.output.decode("ascii")}'
            res = res.decode('utf-8').strip()
            assert 'Completed minimization' in res, 'Failed at system ' + system
            assert res.endswith('Completed simulation!'), 'Failed at system ' + system

if __name__ == "__main__":
    unittest.main(verbosity=2)


import os
import warnings
from pathlib import Path

_htmdconfigfolder = os.path.join(os.path.expanduser("~"), ".htmd")
_warningsfile = os.path.join(_htmdconfigfolder, ".disablewarnings")
try:
    os.makedirs(_htmdconfigfolder, exist_ok=True)
    if not os.path.exists(_warningsfile):
        Path(_warningsfile).touch()
except:
    pass


def _issueWarnings():
    if not os.path.exists(_warningsfile):
        return  # Don't issue warnings if we failed to create the disable file

    if ("CI" in os.environ) and os.environ["CI"]:
        return  # Don't issue warnings if running in CI

    with open(_warningsfile, "r") as f:
        disabledversions = f.read().splitlines()

    if "1.16" not in disabledversions:
        warnings.warn(
            "As of HTMD 1.16 the default number of threads HTMD spawns for calculations is set to 1. "
            "You can enable parallelism at your own risk using `config(njobs=-2)` in the beginning of your scripts. "
            "To disable this warning run once: `from htmd import _disableWarnings; _disableWarnings('1.16');`",
            UserWarning,
        )

    if "1.21" not in disabledversions:
        warnings.warn(
            "As of HTMD 1.21 support for ACEMD v2 has stopped. Please use ACEMD3 instead"
            " as well as the corresponding equilibration and production protocols. "
            "To disable this warning run once: `from htmd import _disableWarnings; _disableWarnings('1.21');`",
            UserWarning,
        )


def _disableWarnings(version):
    if not os.path.exists(_warningsfile):
        return

    with open(_warningsfile, "r") as f:
        versions = f.read().splitlines()

    if version not in versions:
        versions.append(version)

    with open(_warningsfile, "w") as f:
        f.write("\n".join(versions))




from moleculekit.util import *
from moleculekit.dihedral import *
import logging


logger = logging.getLogger(__name__)

logger.warning('Please do not import from htmd.molecule.util. This will be deprecated. '
                'Use moleculekit.util instead or moleculekit.dihedral for the dihedral functionality.')